# -*- coding: utf-8 -*-
#
# Copyright 2014 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls JMD
#
# mpuls JMD is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls JMD is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls JMD. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls JMD has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Ludwig Reiter <ludwig.reiter@intevation.de>
#

import logging
import os.path
import paste
import datetime
from StringIO import StringIO

from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.styles import StyleSheet1
from reportlab.lib.pagesizes import A4, cm
from reportlab.platypus import Paragraph, SimpleDocTemplate, Table
from reportlab.platypus.flowables import KeepTogether, Spacer, Image, \
    CondPageBreak
from reportlab.lib import colors

from pylons import config

from formed.renderer.html import TreeFormData
from mpulsweb.controllers.case import CaseController as MpulsCaseController
from mpulsweb.lib.base import request
from mpulsweb.lib.security import checkRole
from mpulsweb.lib.helpers import getUserFirstname, getUserLastname


log = logging.getLogger(__name__)


styleSheet = StyleSheet1()
styleSheet.add(ParagraphStyle(name='Normal',
                              fontName='Helvetica',
                              fontSize=10,
                              leading=12,
                              spaceBefore=0,
                              spaceAfter=12))
styleSheet.add(ParagraphStyle(name='Textblock',
                              parent=styleSheet['Normal'],
                              spaceAfter=6))
styleSheet.add(ParagraphStyle(name='Goal',
                              fontName='Helvetica',
                              fontSize=10,
                              leading=12,
                              spaceAfter=12,
                              leftIndent=1*cm))
styleSheet.add(ParagraphStyle(name='InvolvedParties',
                              fontName='Helvetica',
                              fontSize=12,
                              leading=14,
                              spaceAfter=12))
styleSheet.add(ParagraphStyle(name='smallestHeader',
                              fontName='Helvetica-Bold',
                              fontSize=11,
                              leading=14,
                              spaceBefore=12,
                              spaceAfter=6,
                              leftIndent=1*cm))
styleSheet.add(ParagraphStyle(name='smallHeader',
                              fontName='Helvetica-Bold',
                              fontSize=12,
                              leading=16,
                              spaceBefore=36,
                              spaceAfter=12))
styleSheet.add(ParagraphStyle(name='Header',
                              fontName='Helvetica-Bold',
                              fontSize=12,
                              leading=24,
                              spaceBefore=24,
                              spaceAfter=12))
styleSheet.add(ParagraphStyle(name='BigHeader',
                              fontName='Helvetica-Bold',
                              fontSize=24,
                              leading=30))
styleSheet.add(ParagraphStyle(parent=styleSheet['Normal'],
                              spaceBefore=40,
                              name="FooterIcon"))


class SupportplanController(MpulsCaseController):

    '''
    Print a support plan specified in mpuls/issue4369 with help of reportlab
    Draft at the moment
    '''

    @checkRole('cm')
    def print_supportplan(self, id):
        case_id = self._checkInt(id)
        case = self._loadCase(case_id, read_only=True)
        instance_tree = case.get_formed_instance_tree()
        tree_form_data = TreeFormData(instance_tree)
        file1 = StringIO()

        doc = SimpleDocTemplate(file1,
                                pagesize=A4,
                                leftMargin=2.5*cm,
                                rightMargin=2*cm,
                                topMargin=1*cm,
                                bottomMargin=1*cm,
                                title=u'Integrationsförderplan',
                                author='impuls JMD')

        elements = []
        rgs = sorted(filter(is_in_process,
                            tree_form_data.get_repeat_group("rg_ziele")),
                     cmp=rg_compare1)

        elements.extend(self._generate_header(instance_tree))
        counter = 1
        previous_goal = ''
        for rg in rgs:
            elements.extend(self._generate_rg_page_part(rg, counter))
            counter += 1
            previous_goal = get_choice_description(rg, "fp_ziel")
        elements.extend(self._generate_footer(case))

        doc.build(elements)

        return self._download(file1.getvalue(), "application/pdf",
                              "Integrationsfoerderplan-{0}".format(case_id),
                              "pdf")

    def _grouping_rgs(self, rgs):
        res1 = []
        res2 = None
        for rg in rgs:
            if not res2:
                res2 = [rg]
            elif (get_leaf_value(res2[-1], "fp_ziel") ==
                  get_leaf_value(rg, "fp_ziel")):
                res2.append(rg)
            else:
                res1.append(res2)
                res2 = [rg]
        if res2:
            res1.append(res2)
        return res1

    def _generate_header(self, instance_tree):
        '''
        Generate header page elements
        '''
        heading = [[p(u'Integrationsförderplan', 'BigHeader'),
                    Image(os.path.join(self._get_static_dir(),
                                       "images", "logos", "header.jpg"),
                          width=6*cm, height=2*cm)
                    ]]
        table_style = [('LEFTPADDING', (0, 0), (- 1, -1), 0),
                       ('VALIGN', (0, 0), (- 1, -1), 'CENTER'),
                       ('ALIGN', (1, 0), (- 1, -1), 'RIGHT')]
        elements = []
        elements.append(Table(heading, hAlign='LEFT',
                              colWidths=(10*cm, 7*cm), style=table_style))

        elements.extend(
            p(u'Jugendliche/r: <strong>{0} {1}</strong>'.format(
                instance_tree.get_node('vorname').getValue(),
                instance_tree.get_node('name').getValue()),
              'InvolvedParties'))
        elements.extend(
            p(u'JMD Berater/in: <strong>{0} {1}</strong>'.format(
                getUserFirstname(),
                getUserLastname()),
              'InvolvedParties'))

        return elements

    def _generate_rg_page_part(self, repeat_group_node, counter):
        elements = []
        elements.extend(self._generate_goal_table(repeat_group_node, counter))
        elements.extend(self._generate_taskpart_table(repeat_group_node))

        # Goals
        has_goal = False
        goals = []
        for goal_nr in range(1, 6):
            if self._generate_concrete_goal_table(repeat_group_node, goal_nr):
                has_goal = True
                goals.extend(self._generate_concrete_goal_table(repeat_group_node, goal_nr))

        if has_goal:
            elements.extend(goals)

        return elements

    def _generate_footer(self, case):
        '''
        Generate the footer with the appropriate logo and final statement
        '''
        elements = []
        elements.append(Spacer(width=10*cm, height=2*cm))
        elements_together = []
        elements_together.extend(p(u"Wir vereinbaren, dass wir zusammenarbeiten"
                                   u" möchten und die Absprachen einhalten.",
                                   'Normal'))
        elements_together.extend(p(u"Ausgedruckt am {0}".format(
            datetime.date.today().strftime("%d.%m.%Y")), 'Normal'))
        elements.append(KeepTogether(elements_together))

        return elements

    def _generate_goal_table(self, repeat_group_node, counter):
        '''
        Generate the goal description elements
        '''
        fp_ziel = get_choice_description(repeat_group_node, "fp_ziel")
        elements = []
        heading = p(u'{0}. Förderziel - {1}'.format(counter, fp_ziel),
                    "smallHeader")[0]
        w, h = heading.wrap(0, 0)
        elements.append(CondPageBreak(h * 1.1))
        elements.append(heading)
        elements.extend(p(get_leaf_value_or_na(repeat_group_node,
                                               "fp_ziel_beschreibung")))

        return elements

    def _generate_taskpart_table(self, repeat_group):
        '''
        Generate the task description elements
        '''
        elements = []
        elements_together = []

        elements_together.extend(
            p(u'<strong>Leistungen der/des Teilnehmenden:</strong> {0}'.format(
                get_leaf_value_or_na(repeat_group, "fp_leistung_teilnehmer"))))

        elements_together.extend(
            p(u'<strong>Leistung des JMD:</strong> {0}'.format(
                get_leaf_value_or_na(repeat_group, "fp_leistung_jmd"))))

        elements_together.extend(
            p(u'Dieses Ziel wurde am {0} vereinbart und soll '
              'bis zum {1} umgesetzt werden.'.format(
                  get_leaf_value_or_na(repeat_group, "fp_datum_vereinbarung"),
                  get_leaf_value_or_na(repeat_group, "fp_datum_umsetzung"))))
        table_style = []
        table_style.append(('LEFTPADDING', (0, 0), (- 1, -1), 0))
        table_style.append(('BOTTOMPADDING', (0, 0), (- 1, -1), 0))
        elements.append(KeepTogether(elements_together))

        return elements

    def _generate_concrete_goal_table(self, repeat_group, goal_nr=1):
        '''
        Generate the sub goal given

        If the goal has no description, it is considered as not set and
        the function will return None. Otherwise the reportlab elements
        are returned.
        '''

        map = {'ziel': 'fp_uziel{0}'.format(goal_nr),
               'ziel_erreicht': 'fp_uziel{0}_erreicht'.format(goal_nr),
               'angebot': 'fp_uziel{0}_angebot'.format(goal_nr),
               'art': 'ua_uziel{0}_art'.format(goal_nr),
               'datum_start': 'ua_uziel{0}_datum_start'.format(goal_nr),
               'datum_ende': 'ua_uziel{0}_datum_ende'.format(goal_nr),
               'einrichtung': 'fp_externe_einrichtung_{0}_name'.format(goal_nr),
               'einrichtung_datum':
               'fp_externe_einrichtung_{0}_datum'.format(goal_nr),
               }
        if not get_leaf_value(repeat_group, map["ziel"]):
            return None

        elements = []
        elements_together = []

        elements_together.extend(p(u'Unterziel {0}'.format(goal_nr),
                                   'smallestHeader'))
        elements_together.extend(p(get_leaf_value_or_na(repeat_group,
                                                        map["ziel"]), 'Goal'))
        elements_together.append(table_unterstuetzungsangebot(repeat_group,
                                                              map))

        elements.append(KeepTogether(elements_together))
        return elements

    def _get_static_dir(self):
        return config['pylons.paths']['static_files'][0]

    def _download(self, data, mime_type, name, extension, **extra):
        disposition = 'attachment; filename="%s.%s"' % (name, extension)
        data = paste.fileapp.DataApp(data,
                                     content_type=mime_type,
                                     content_disposition=disposition,
                                     cache_control='must-revalidate',
                                     pragma='public',
                                     **extra)
        return data(request.environ, self.start_response)


def p(content, stylesheet='Textblock'):
    '''
    Small helper function that simply returns a paragraph with the given content
    '''
    paragraphs = []
    if content.__class__.__name__ == 'tuple':
        items = content
    else:
        items = (content,)

    # TODO: Actually, this doesn't work. Paragraph() ignores the "<para>" tags.
    # Everything is in one para block.
    for item in items:
        for p in item.split('\r\n'):
            if len(p.strip()) > 0:
                paragraphs.append(Paragraph(u'<para>{0}</para>'.format(p),
                                            styleSheet[stylesheet]))

    return paragraphs

def table_unterstuetzungsangebot(repeat_group, map):
    '''
    Small helper function that returns a table for a "Unterstützungsangebot".
    '''
    table_style = []
    table_style.append(('BOX', (0, 0), (-1, -1), 0.25, colors.black))
    table_style.append(('INNERGRID', (0, 0), (- 1, -1), 0.25, colors.black))
    table_style.append(('VALIGN', (0, 0), (- 1, -1), 'TOP'))

    data = []
    data.append([p(u"Erreicht"),
                 get_choice_description(repeat_group, map["ziel_erreicht"])])
    data.append([p(u"Unterstützungsangebot vorhanden"),
                 get_choice_description(repeat_group, map['angebot'])])

    if get_leaf_value(repeat_group, map['angebot']) == 1:

        data.append([p(u"Art der Unterstützung"),
                     p(get_leaf_value_or_na(repeat_group, map['art']))])

        data.append([p(u"Beginn und voraussichtliches Ende "),
                     u"{0} - {1}".format(
                         get_leaf_value_or_na(repeat_group, map['datum_start']),
                         get_leaf_value_or_na(repeat_group,
                                              map['datum_ende']))])

        data.append([p(u"Externe Einrichtung"),
                     p(get_leaf_value_or_na(repeat_group,
                                            map['einrichtung'])),])
        data.append([p(u"Einbeziehung bis"),
                     get_leaf_value_or_na(repeat_group,
                                          map['einrichtung_datum'])])

    return Table(data,
                 colWidths=(6.5*cm, 8.5*cm), hAlign='RIGHT', style=table_style)


def get_leaf_value(repeat_group_node, fieldname):
    '''
    Returns the value of the leaf
    '''
    return repeat_group_node.get_node(fieldname).getValue()


def get_leaf_value_or_na(repeat_group_node, fieldname, na='Keine Angabe'):
    '''
    Returns the value of the leaf.

    Differences to get_leaf_value:
        * Dates are formatted according to current locale
        * If the value of the leaf is None, returns the string of 'na'
    '''
    value = get_leaf_value(repeat_group_node, fieldname)
    if value.__class__.__name__ == 'date':
        value = value.strftime('%x')
    return value or na


def get_choice_description(repeat_group_node, fieldname):
    '''
    Returns the option value as shown in the application's drop down list
    '''
    node = repeat_group_node.get_node(fieldname)
    meta = node.getMeta()
    mapping = meta.getValueMapping()
    if mapping:
        return mapping.get(node.getValue())
    return None


def is_in_process(rg):
    '''
    Check a rg_ziele rg to be in process.
    '''
    if get_leaf_value(rg, "fp_ziel_erreicht") == 2:
        return True
    else:
        return False


def rg_compare1(rg1, rg2):
    a1 = get_leaf_value(rg1, "fp_ziel")
    a2 = get_leaf_value(rg2, "fp_ziel")

    if a1 == -1:
        a1 = 1000
    if a2 == -1:
        a2 = 1000
    if a1 == a2:
        return rg_compare2(rg1, rg2)
    else:
        return a1 - a2


def rg_compare2(rg1, rg2):
    d1 = get_leaf_value(rg1, "fp_datum_vereinbarung")
    d2 = get_leaf_value(rg2, "fp_datum_vereinbarung")

    if not d1:
        d1 = datetime.date(1975, 1, 1)
    if not d2:
        d2 = datetime.date(1975, 1, 1)

    if d1 < d2:
        return 1
    if d1 > d2:
        return -1
    return rg_compare3(rg1, rg2)


def rg_compare3(rg1, rg2):
    d1 = get_leaf_value(rg1, "fp_datum_umsetzung")
    d2 = get_leaf_value(rg2, "fp_datum_umsetzung")

    if not d1:
        d1 = datetime.date(1975, 1, 1)
    if not d2:
        d2 = datetime.date(1975, 1, 1)

    if d1 < d2:
        return 1
    if d1 > d2:
        return -1
    return 0
