# -*- coding: utf-8 -*-
#
# Copyright 2007-2018 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Ludwig Reiter (ludwig.reiter@intevation.de)
# Dirk Massmann (dirk.massmann@intevation.de)

import logging
import datetime

from mpulsweb.lib.base import session, g
from mpulsweb.lib.db import db
from mpulsweb.lib.translation import _
import mpulsweb.lib.searchexpr as S
from mpulsweb.model.appointment import AppointmentOverview, \
        CaseAppointmentFactory, MaxSaveTimeReminderOverview
from mpulsweb.model.case import ExtendedCaseOverview


log = logging.getLogger(__name__)


class AgeReminderOverview(AppointmentOverview):
    """Generic overview controlling age of person."""

    AGE_QUERY_TEMPLATE = """
    SELECT m.id
    FROM master_tbl_view m
    WHERE m.bearbeiter_id = %(editor)s
    AND {CONDITION}
    """
    AGE_CONDITION = ""
    AGE_MESSAGE = ""
    AGE_MESSAGE_TITLE = ""

    def __init__(self, case=None):
        self.create_appointments(self.search_cases(case))

    def get_query(self, case=None):
        conditions = [self.AGE_CONDITION]
        params = {'editor': self.getUserId()}
        if case is not None:
            conditions.append("m.id = %(case_id)s")
            params["case_id"] = case.id
        condition = " AND ".join(conditions)
        return (self.AGE_QUERY_TEMPLATE.format(CONDITION=condition), params)

    def search_cases(self, case=None):
        with db.cursor() as cur:
            query, params = self.get_query(case)
            cur.execute(query, params)
            case_ids = set(int(r[0]) for r in cur.fetchall())

        sm = g.case_factory.get_state_machine()
        search_phases = sm.get_search_item_ids()

        # Handle some special cases that let us avoid additional
        # database queries.
        if not case_ids:
            # No reminders at all, so no cases
            return []
        if case is not None:
            # There must be exactly one potential case with a reminder:
            # case_ids is true here and a case was given as argument.
            # That case must be the one found.
            assert set([case.id]) == case_ids

            # The case only actually has a reminder if it's in the right
            # state and phase
            state = case.getState()
            if state.getState() in [1, 2] and state.getPhase() in search_phases:
                return [case]
            else:
                return []

        cases = ExtendedCaseOverview()
        cases.search(S.And([S.Status([1, 2]),
                            S.Choice("phase", search_phases),
                            S.Choice("id", list(case_ids))]))
        return [case for case in cases.getDatasets()
                     if case.id in case_ids]

    def create_appointments(self, case_list):
        self.appointment_list = [self.makeAppointment(case,
                                                      self.AGE_MESSAGE,
                                                      self.AGE_MESSAGE_TITLE)
                                 for case in case_list]

    def makeAppointment(self, case, description, title):
        appointment_type = 2
        return CaseAppointmentFactory().createByData(None,
                                                     title,
                                                     datetime.datetime.now(),
                                                     None,
                                                     description,
                                                     case.id,
                                                     appointment_type)

    def getUserId(self):
        """
        Return the User-ID. This function can be overwritten by applications
        that have a "structure" part where the user that owns the cases is not
        the user logged in.
        """
        return session['USER_AUTHORIZED'].id


class MinAgeReminderOverview(AgeReminderOverview):
    """JMD specific overview. If a young person hasn't reached an age of
    12 years, the cm should be informed."""

    AGE_CONDITION = """age(m.geburtsdatum) < interval '12 years'"""
    AGE_MESSAGE = _('The age of the young person is under 12.')
    AGE_MESSAGE_TITLE = _('Age under 12')


class MaxAgeReminderOverview(AgeReminderOverview):
    """JMD specific overview. If a young person reaches an age of 27, the cm
    should be informed."""

    AGE_CONDITION = """age(m.geburtsdatum) > interval '27 years'"""
    AGE_MESSAGE = _('The age of the young person is above 27.')
    AGE_MESSAGE_TITLE = _('Age above 27')


class JustBeforeMaxAgeReminderOverview(AgeReminderOverview):
    """JMD specific overview. If a young person reaches an age of 27 within the
    next 6 months, the cm should be informed."""

    AGE_CONDITION = """age(m.geburtsdatum) > interval '318 months'
                   AND age(m.geburtsdatum) < interval '324 months'"""
    AGE_MESSAGE = _('The age of the young person will be 27'
                    ' within the next 6 months.')
    AGE_MESSAGE_TITLE = _('Reaching age of 27 soon')


def automatic_reminders_for_case(case):
    return (MinAgeReminderOverview(case).getDatasets()
            + JustBeforeMaxAgeReminderOverview(case).getDatasets()
            + MaxAgeReminderOverview(case).getDatasets()
            + MaxSaveTimeReminderOverview(case).getDatasets())
