# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

import logging
import formencode
from datetime import datetime, date
from collections import Counter

import paste

from pylons import app_globals as g
from pylons.controllers.util import abort, redirect

from mpulsweb.lib.translation import _
from mpulsweb.lib.db import db
from mpulsweb.lib.security import checkRole
from mpulsweb.model.logbook import (
    Logbook,
    LogbookEntry,
    build_logbook_entry,
    export_as_xlsx
)
from mpulsweb.model.document import Document, LogbookDocumentAPI, \
    DocumentNotFoundException
from mpulsweb.lib.helpers import url
from mpulsweb.lib.base import DownloadController, c, render, request, session, \
     render_htmlfill, restrict_to_POST, abort_unless_POST
from mpulsweb.lib.validators import CreateLogbookForm, EditLogbookForm, \
     LogbookCaseDocumentForm, RenameLogbookDocumentForm


log = logging.getLogger(__name__)


def get_sort(params):
    sort = params.get('sort_field', 'id')
    order = params.get('sort_order', 'asc')
    return sort, order


class LogbookController(DownloadController):

    def _prepare_defaults(self, entry):
        """Prepare the dictionary used by htmlfill"""
        entry_data = {}
        entry_data['id'] = entry.id
        entry_data['date'] = entry.getDate()
        entry_data['time'] = entry.getTime()
        entry_data['duration'] = entry.getDuration()
        entry_data['art'] = entry.getKindAsInt()
        entry_data['notice'] = entry.getNotice()
        entry_data['typ'] = entry.getTyp()
        entry_data['short_notice'] = entry.getShortNotice()

        return entry_data

    @checkRole(('cm'))
    def index(self, id):
        return self.overview(id)

    def overview(self, id):
        sort, order = get_sort(request.params)
        id = self._checkInt(id)

        ## Load logbook
        logbook = Logbook()
        logbook.loadById(id, sort, order)

        self._ensure_session_case(id)

        c.logbook = logbook
        c.logbook_documents = LogbookDocumentAPI.list_case_logbook_documents(id)
        c.logbook_document_counts = Counter(doc.logbook_id
                                            for doc in c.logbook_documents)

        ## Build evaluations
        c.eval_effort = logbook.getEvaluation()
        c.eval_typ_effort = logbook.getTypEvaluation()

        return render('/logbook/overview.mako')

    @checkRole(('cm'))
    def show(self, id):
        id = self._checkInt(id)
        try:
            entry = LogbookEntry()
            entry.loadById(id)
        except Exception, e:
            c.dialog_title = _(u"Showing of logbook entry failed!")
            c.dialog_text = _(u"An error occured while showing the logbook"
                              u" entry.")
            c.url_ok = url(controller='logbook', action='index',
                           id=session['case'].id)
            log.exception(e)
            return render('/logbook/dialogs/failed_show_logbook_entry.mako')

        self._ensure_session_case(entry.getMasterId())
        c.entry = entry
        c.files = LogbookDocumentAPI(id).list_db()
        return render('/logbook/show.mako')

    @checkRole('cm')
    def new(self, id):
        id = self._checkInt(id)
        self._ensure_session_case(id)
        if not session["case"].is_editable_state():
            abort(401)

        entry = LogbookEntry()
        entry.setDate(datetime.now())
        entry.setKind(-1)

        form_defaults = {}
        c.form_errors = {}

        # Set the default values, also for all the "types" from the JSON file
        form_defaults['date'] = entry.getDate()
        form_defaults['time'] = entry.getTime()
        form_defaults['kind'] = entry.getKindAsInt()
        form_defaults['case_id'] = id
        for num, typ in enumerate(g.mpuls_config.get('logbook', 'types')):
            form_defaults[typ.get('db_field')] = -1  #only if it's a drop-down!

        form = render('/logbook/new.mako')
        return render_htmlfill(form, defaults=form_defaults,
                               auto_insert_errors=False, errors=c.form_errors)

    @checkRole('cm')
    @restrict_to_POST
    def newAction(self):
        validator = CreateLogbookForm(
            max_size=g.mpuls_config.get("case", "max-document-size"),
            max_size_text=g.mpuls_config.get("case", "max-document-size-text")
            )
        c.form_errors = {}
        c.form_result = {}

        case_id = self._checkInt(request.params.get("case_id"))
        if case_id is not None:
            self._ensure_session_case(case_id)
            if not session["case"].is_editable_state():
                abort(401)

        try:
            form_result = validator.to_python(request.params)
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            form = render('/logbook/new.mako')
            return render_htmlfill(form, defaults=c.form_result,
                                   auto_insert_errors=False,
                                   errors=c.form_errors)

        try:
            log.debug("New Logbook entry from case %s", case_id)
            logbook = Logbook(case_id)
            entry = build_logbook_entry(kind=form_result["art"],
                                        date=form_result["date"],
                                        time=form_result["time"],
                                        duration=form_result.get("duration"),
                                        notice=form_result.get("notice"),
                                        short_notice=
                                        form_result.get("short_notice"),
                                        typ=form_result.get("typ"))
            logbook.addEntry(entry)
            with db.transaction() as cur:
                logbook.store_cursor(cur, session['USER_AUTHORIZED'].id)
                uploaded = form_result.get('file')
                if uploaded is not None:
                    api = LogbookDocumentAPI(entry.id)
                    api.upload_documents(cur, uploaded)

            c.url_ok = url(controller='logbook', action='index', id=case_id)
            c.success_for = _(u"Entry successfully created.")
            c.success_text = _(u"The logbook entry was sucessfully created")
            if request.params.get('save_and_back'):
                return render('/logbook/dialogs/success_create_logbook_entry.mako')
            elif request.params.get('save_and_new'):
                return self.new(case_id)
            else:
                entry_data = self._prepare_defaults(entry)
                c.entry = entry
                c.form_errors = {}
                form = render('/logbook/edit.mako')
                return render_htmlfill(form, defaults=entry_data,
                                   auto_insert_errors=False,
                                   errors={})
        except Exception, e:
            log.exception("Create new logbook entry failed")
            c.dialog_title = _(u"Logbook entry was not created!")
            c.dialog_text = _(u"An error occured while creating a logbook"
                              u" entry")
            c.url_ok = url(controller='logbook', action='index', id=case_id)
            return render('/logbook/dialogs/failed_create_logbook_entry.mako')


    @checkRole('cm')
    def edit(self, id):
        id = self._checkInt(id)
        c.form_errors = {}
        try:
            entry = LogbookEntry()
            entry.loadById(id)
            self._ensure_session_case(entry.getMasterId())
            if not entry.is_editable_in_case(session["case"]):
                abort(401)

            entry_data = self._prepare_defaults(entry)
            c.entry = entry
            form = render('/logbook/edit.mako')
            return render_htmlfill(form, defaults=entry_data,
                                   auto_insert_errors=False, errors={})
        except Exception, e:
            log.exception(e)
            c.dialog_title = _(u"Showing of logbook entry failed!")
            c.dialog_text = _(u"An error occured while showing the logbook"
                              u" entry.")
            c.url_ok = url(controller='logbook', action='index',
                           id=session['case'].id)
            return render('/logbook/dialogs/failed_show_logbook_entry.mako')

    @checkRole('cm')
    @restrict_to_POST
    def editAction(self):
        params = request.params
        id = self._checkInt(params.getone('id'))

        # Abort if editing not allowed. As a side effect it makes sure
        # that session["case"] is set correctly.
        entry = LogbookEntry()
        entry.loadById(id)
        self._ensure_session_case(entry.getMasterId())
        if not entry.is_editable_in_case(session["case"]):
            abort(401)

        validator = EditLogbookForm()

        try:
            form_result = validator.to_python(params)
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            c.entry = entry
            form = render('/logbook/edit.mako')
            return render_htmlfill(form, defaults=c.form_result,
                                   auto_insert_errors=False,
                                   errors=c.form_errors)

        try:
            entry = build_logbook_entry(kind=form_result["art"],
                                        date=form_result["date"],
                                        time=form_result["time"],
                                        duration=form_result.get("duration"),
                                        notice=form_result.get("notice"),
                                        short_notice=
                                        form_result.get("short_notice"),
                                        typ=form_result.get("typ"))
            entry.setId(id)
            entry.storeForUser(session['USER_AUTHORIZED'].id)

            c.success_for = _(u"Changes on logbook successfull!")
            c.success_text = _(u"The changes on the logbook entry were"
                               u" applied successfully.")
            c.url_ok = url(controller='logbook', action='index',
                           id=session['case'].id)
            if params.get('save_and_back'):
                return render('/logbook/dialogs/success_save_logbook_entry.mako')
            elif params.get('save_and_new'):
                return self.new(session['case'].id)
            else:
                entry_data = self._prepare_defaults(entry)
                c.entry = entry
                c.form_errors = {}
                form = render('/logbook/edit.mako')
                return render_htmlfill(form, defaults=entry_data,
                                   auto_insert_errors=False,
                                   errors={})
        except Exception, e:
            log.exception(e)
            c.dialog_title = _(u"Changes on logbook failed!")
            c.dialog_text = _(u"An error occured applying changes on the"
                              u" logbook entry")
            c.url_ok = url(controller='logbook', action='index',
                           id=session['case'].id)
            return render('/logbook/dialogs/failed_save_logbook_entry.mako')

    @checkRole('cm')
    def delete(self, id, confirmed=False):
        case_id = None
        id = self._checkInt(id)
        try:
            entry = LogbookEntry()
            entry.loadById(id)
            case_id = entry.getMasterId()
            self._ensure_session_case(case_id)
        except Exception, e:
            log.exception(e)
            c.dialog_title = _(u"Checking of logbook entry delete permission"
                             u" failed!")
            c.dialog_text = _(u"An error occured while checking delete"
                              u" permission for the logbook entry.")
            c.url_ok = url(controller='logbook', action='index',
                           id=(case_id if case_id is not None
                               else session["case"].id))
            return render('/logbook/dialogs/failed_delete_logbook_entry.mako')

        if not entry.is_editable_in_case(session["case"]):
            # The case is either not editable or the entry to be deleted
            # is an automatic entry. In both cases, deletion is not
            # permitted
            abort(401)

        confirmed = self._checkBool(confirmed)
        if confirmed == 1:
            abort_unless_POST()
            try:
                Logbook().deleteEntryById(int(id))
            except Exception, e:
                log.exception(e)
                c.dialog_title = _(u"Deletion of logbook entry failed!")
                c.dialog_text = _(u"An error occured while deleting the"
                                  u" logbook entry.")
                c.url_ok = url(controller='logbook', action='index', id=case_id)
                return render('/logbook/dialogs/failed_delete_logbook_entry.mako')
            return self._delete_succeded(case_id)
        else:
            c.confirm_for = _(u"Delete logbook entry?")
            c.question = _(u"Do you want to delete this logbook entry?\n"
                           u"The data will be irrevocably lost.")
            c.url_yes = url(controller='logbook', action='delete', id=id,
                            confirmed=1)
            c.url_no = url(controller='logbook', action='index', id=case_id)
            return render('/logbook/dialogs/confirm_delete_logbook_entry.mako')

    def _delete_succeded(self, case_id):
        c.success_for = _(u"Logbook entry successfully deleted!")
        c.success_text = _(u"Logbook entry was successfully deleted!")
        c.url_ok = url(controller='logbook', action='index', id=case_id)
        return render('/logbook/dialogs/success_delete_logbook_entry.mako')

    @checkRole('cm')
    def newDocument(self, id):
        id = self._checkInt(id)
        c.entry = LogbookEntry()
        c.entry.loadById(id)
        self._ensure_session_case(c.entry.getMasterId())
        return render('/logbook/newdocument.mako')

    @checkRole('cm')
    @restrict_to_POST
    def newDocumentAction(self):
        logbook_id = self._checkInt(request.params.get("logbook"))
        c.entry = LogbookEntry()
        c.entry.loadById(logbook_id)
        self._ensure_session_case(c.entry.getMasterId())
        if not session["case"].is_editable_state():
            abort(401)

        validator = LogbookCaseDocumentForm(
            max_size=g.mpuls_config.get("case", "max-document-size"),
            max_size_text=g.mpuls_config.get("case", "max-document-size-text")
            )
        try:
            result = validator.to_python(request.params)
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            log.debug(c.form_result)
            log.warn("Invalid form data for logbook document %s"
                     % c.form_errors)
            form = render('/logbook/newdocument.mako')
            return render_htmlfill(form, defaults=c.form_result,
                                   auto_insert_errors=False,
                                   errors=c.form_errors)

        logbook_id = result['logbook']
        api = LogbookDocumentAPI(logbook_id)
        api.upload_documents_transaction(result['file'])

        c.dialog_title = _(u"Attachment added!")
        c.dialog_text = _(u"Please click OK to continue.")
        c.url_ok = url(controller="logbook", action="show", id=logbook_id)
        return render('logbook/dialogs/success_upload_document.mako')


    @checkRole(('admin', 'cm'))
    def showDocument(self, id):
        id = self._checkInt(id)
        api = LogbookDocumentAPI(None)
        try:
            document = api.load_db(id, load_contents_immediately=True)
        except DocumentNotFoundException:
            # the document could not be found, so we should send a 404
            # response to the client.
            abort(404)

        data = paste.fileapp.DataApp(document.data,
                                     content_type=document.mime,
                                     cache_control='must-revalidate',
                                     pragma='public')
        return data(request.environ, self.start_response)


    def _load_editable_document(self, doc_id):
        doc_id = self._checkInt(doc_id)
        api = LogbookDocumentAPI(None)
        try:
            doc = api.load_db(doc_id, load_contents_immediately=False)
        except DocumentNotFoundException:
            abort(404)

        # Abort if editing not allowed. As a side effect it makes sure
        # that session["case"] is set correctly.
        entry = LogbookEntry()
        entry.loadById(doc.logbook_id)
        self._ensure_session_case(doc.case)
        if not entry.is_editable_in_case(session["case"]):
            abort(401)

        c.entry = entry
        return doc

    @checkRole('cm')
    def deleteDocument(self, id, confirmed):
        doc = self._load_editable_document(id)

        confirmed = self._checkBool(confirmed)
        if confirmed == 1:
            abort_unless_POST()
            LogbookDocumentAPI(doc.logbook_id).delete_transaction(doc.id)
            c.url_ok = url(controller="logbook", action="show",
                           id=doc.logbook_id)
            c.success_for = _(u"Attachment deleted!")
            c.success_text = _(u"Please click OK to continue.")
            return render('/logbook/dialogs/'
                          'success_delete_document.mako')
        else:
            c.context = "../main.mako"
            c.confirm_for = _(u"Delete attachment?")
            c.question = _(u"Do you want to irrevocably delete the attachment"
                           u" '%s'?") % doc.name
            c.url_yes = url(controller='logbook', action='deleteDocument',
                            id=id, confirmed=1)
            c.url_no = url(controller="logbook", action="show",
                           id=doc.logbook_id)
            return render('/logbook/dialogs/'
                          'confirm_delete_document.mako')

    @checkRole('cm')
    def renameDocument(self, id):
        doc = self._load_editable_document(id)

        c.id = doc.id
        c.logbook_id = doc.logbook_id
        c.name = doc.getName()
        if not c.name:
            c.name = ""

        return render('/logbook/renamedocument.mako')

    @checkRole('cm')
    @restrict_to_POST
    def renameDocumentAction(self, id):
        doc = self._load_editable_document(id)
        validator = RenameLogbookDocumentForm()
        try:
            result = validator.to_python(request.params)
        except formencode.Invalid, error:
            c.id = doc.id
            c.name = doc.getName()
            if not c.name:
                c.name = ""
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            log.debug(c.form_result)
            form = render('/logbook/renamedocument.mako')
            return render_htmlfill(form, defaults=c.form_result,
                                   auto_insert_errors=False,
                                   errors=c.form_errors)

        api = LogbookDocumentAPI(doc.logbook_id)
        api.rename_transaction(doc.id, result['name'])

        redirect(url(controller="logbook", action="show", id=doc.logbook_id))

    @checkRole(('cm'))
    def printout(self, id):
        id = self._checkInt(id)
        c.print_version = 1

        # Load logbook
        logbook = Logbook()
        logbook.loadById(id)
        c.logbook = logbook

        # Load case
        case = self._loadCase(logbook.getMasterId(), read_only=True)
        c.case = case

        return render('/logbook/overview_print.mako')

    @checkRole(('cm',))
    def export(self, id):
        id = self._checkInt(id)

        # Load logbook for id
        logbook = Logbook()
        logbook.loadById(id)

        case = self._loadCase(logbook.getMasterId(), read_only=True)

        # Work happens in export_as_xls
        return self._download(export_as_xlsx(logbook, case),
                              ("application/vnd.openxmlformats-"
                               "officedocument.spreadsheetml.sheet"),
                              "tagebuch-export", extension="xlsx")
