# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

"""The application's Globals object"""

import os
import logging
import ConfigParser

from paste.reloader import watch_file
from paste.registry import StackedObjectProxy

from formed.instance.backends.postgres import DBFactory as InstanceFactory
from formed.meta.io.document import openDocument
from formed.meta.structure import StructureTree
from formed.meta.conditionals import FormInfo

from libmpuls.service.tags import TagSetterConfig

from mpulsweb.config.importer import import_overridable_module
from mpulsweb.lib.config import MpulsAppConfig, MpulsDBMapping, MpulsDBList
from mpulsweb.lib.db import PostgresDBInterface
from mpulsweb.lib.login_db import LoginDatabase
from mpulsweb.model.annotations import AnnotationsProvider

log = logging.getLogger(__name__)


# MPuls session object.  Application specific wrapper around the pylons
# session object.  The actual functionality is defined in
# mpulsweb/lib/session.py.  It's set on a per-request basis in
# mpulsweb/lib/base.py.  It's located here in app_globals.py to avoid
# excessive problems with circular imports.
session = StackedObjectProxy()


class Globals(object):

    """Globals acts as a container for objects available throughout the
    life of the application
    """

    def __init__(self, config):
        """One instance of Globals is created during application
        initialization and is available during requests via the 'g'
        variable
        """
        def filename_from_config(setting, default):
            return os.path.join(config.get('pylons.paths').get('root'),
                                '..', config.get(setting, default))

        # Load mpuls configuration
        config_file = filename_from_config('mpuls.app.config',
                                           'default_config.json')
        mapping_file = filename_from_config('mpuls.app.dbmapping',
                                            'default_dbmapping.json')
        tagging_file = filename_from_config('mpuls.app.tagging',
                                            'default_tagging.xml')
        dblist_file = filename_from_config('mpuls.app.dblist',
                                           'default_dblist.json')

        # Load ini file and make all sections available a global var.
        self.mpuls_ini = ConfigParser.ConfigParser()
        path = os.path.join(config['here'], config['__file__'])
        self.mpuls_ini.readfp(open(path))

        self.config_path = os.path.dirname(config_file)
        self.mpuls_config = self._get_config(config_file)
        self.mpuls_dbmapping = MpulsDBMapping(mapping_file)
        self.mpuls_dblist = MpulsDBList(dblist_file)

        self._load_form(config)

        # Load taggingconfig
        log.debug("Loading configuration for tagging from %s" % tagging_file)
        self.taggingConfFile = tagging_file
        self.taggingConf = TagSetterConfig(tagging_file)

        # tell paste to restart the server if the formed files change
        watch_file(tagging_file)

        if self.mpuls_config.is_enabled("security", "display-last-logins"):
            login_db_path = config.get("mpuls.app.login_db")
            if not login_db_path:
                raise Exception("You need to set mpuls.app.login_db,"
                                "if display-last-logins is set.")
            self.login_db = LoginDatabase(login_db_path)
            self.login_db.check()
        else:
            self.login_db = None

        if config.get("debug", False):
            import warnings
            warnings.simplefilter('default')

        # Running as an evaluation server?
        self.is_evaluation_server = config.get('jmd.eval.server', False)

        # TODO: These setups should be while application intialisation, as they
        # are not related to any global vars or functions. Move to
        # config/environment.py

        self._init_case(config)

    def _init_case(self, config):
        from mpulsweb.model.agency import MpulsAgencyFactory, MpulsAgency
        from mpulsweb.model.case import MpulsCase, MpulsCaseFactory
        """Set the mpuls application specific case classes"""
        try:
            case_module = import_overridable_module("model.case", config)
            Case = case_module.Case
            self.case = Case
            CaseFactory = case_module.CaseFactory
            self.case_factory = CaseFactory(self.case)
        except Exception, e:
            # it would be better to not catch all exceptions here
            # an application should eventually not start if a case module
            # couldn't be loaded. At least log the reason as error
            log.exception("Error loading 'model.case' from overridable module."
                          "Taking case from base")
            self.case = MpulsCase
            self.case_factory = MpulsCaseFactory(self.case)

        self.agency = MpulsAgency

        self.case.set_fields(self.mpuls_config.get("case", "fields"))
        self.agency.set_fields(self.mpuls_config.get("agency", "fields"))

        self.agency_factory = MpulsAgencyFactory(self.agency)

        search_module = import_overridable_module("lib.search", config)
        self.case_search = search_module.CaseSearch(
            self.case_factory.get_search_retrieve_fields(),
            self.case_factory.get_search_match_fields())
        self.extended_case_search = search_module.ExtendedCaseSearch(
            self.case_factory.get_search_retrieve_fields())
        self.casecount_search = search_module.CaseCountSearch(
            self.case_factory.get_search_match_fields())
        self.eval_search = search_module.EvaluationSearch(
            self.agency_factory.get_search_retrieve_fields(),
            self.agency_factory.get_search_match_fields())

    def get_instance_factory(self):
        return InstanceFactory(self.formedTree, PostgresDBInterface())

    def get_navigation(self, id, openfolders=None):
        # We import mpulsweb.lib.navigation because doing it at the
        # module-level leads to problems with circular imports with
        # mpulsweb.lib.app_globals.session
        from mpulsweb.lib.navigation import get_navigation
        return get_navigation(id, openfolders=openfolders)

    def _get_config(self, filename):
        return MpulsAppConfig(filename)

    def _formed_file_name(self, config, section, key):
        """Determine the name of a formed related file from config settings."""
        static_dir = config['pylons.paths']['static_files'][0]
        return os.path.join(config.get('mpuls.app.path.formed',
                                       os.path.join(static_dir, 'formed')),
                            self.mpuls_config.get(section, key))

    def _load_form(self, config):
        # Load formed from defined location. If no location is defined. Take
        # the default which is the formed dir under public.
        formed_file = self._formed_file_name(config, 'formed', 'treefile')
        self.formedTree = self._load_formedtree_from_file(formed_file)
        watch_file(formed_file)

        self.structureTree = StructureTree(self.formedTree)

        annotation_file = self._formed_file_name(config, 'formed', 'annotation')
        self.helpData = AnnotationsProvider(annotation_file)
        watch_file(annotation_file)

        self.conditional_form_info = FormInfo(self.formedTree)
        self.conditional_form_info_js = self.conditional_form_info.as_js()

    def _load_formedtree_from_file(self, formed_file):
        """Load a formed tree from a file and return it.
        Derived classes may extend or overwrite this method in order to
        influence how formed trees are loaded so that e.g. rules are
        automatically generated.
        """
        log.debug("loading formed definition from %r", formed_file)
        return openDocument(formed_file)

    def before_hook(self):
        """Called by the base controller from the __before__ method.

        Derived applications can overwrite this method to do additional
        checks before the request is handled.
        """
