# -*- coding: utf-8 -*-
#
# (c) 2014 by Intevation GmbH
# This is Free software under the GPLv3.
# See LICENSE comming with the source of 'mpuls'
# for details.
#
# author: Bjoern Ricks <bjoern.ricks@intevation.de>
#

import logging
import optparse
import os
import subprocess
import sys
import tempfile

command_map = dict()


def register_command(name, command):
    """
    Registers a command callable with name
    """
    command_map[name] = command


def find_command(name):
    """
    Returns the command for name
    """
    return command_map.get(name)


def list_commands():
    """
    Returns a sorted key value list of all commands
    """
    return sorted(command_map.items())


def update_commands(cmd_map):
    """
    Updates the global command map with cmd_map
    """
    command_map.update(cmd_map)


def exec_process(cmd_list, error_msg=None, stderr=None, stdout=None,
                 stdin=None, cwd=None):
    """
    Executes a subprocess specified in cmd_list
    """
    retcode = subprocess.call(cmd_list, stdout=stdout, stderr=stderr,
                              stdin=stdin, cwd=cwd)

    if retcode != 0:
        if error_msg:
            logging.error(error_msg)
        raise Exception(repr(cmd_list) + " failed")


def exec_sql(sql_string, db_name=None, port=None, error_msg=None, cwd=None):
    """
    Executes a psql process for running the sql statement in sql_string
    """
    cmd = ["psql"]

    if db_name:
        cmd.append(db_name)
    if port:
        cmd.append("-p")
        cmd.append(port)

    input = tempfile.TemporaryFile()
    input.write(sql_string)
    input.seek(0)

    exec_process(cmd, stdin=input, error_msg=error_msg, cwd=cwd)


def help_command_list(scriptname):
    print >>sys.stderr, "%s supports the following commands:" % (scriptname,)
    for name, cmd_class in list_commands():
        print >>sys.stderr, "    %s" % name


class CommandMetaClass(type):

    """
    MetaClass to register a command
    """

    def __new__(mcs, name, bases, dict):
        cls = type.__new__(mcs, name, bases, dict)
        if cls.name is not None:
            register_command(cls.name, cls())
        return cls


class BaseCommand(object):

    __metaclass__ = CommandMetaClass

    name = None

    def get_option_parser(self, prog_name):
        """
        Returns the option parser
        """
        parser = optparse.OptionParser(prog="%s %s" % (prog_name, self.name))
        self.add_options(parser)
        return parser

    def add_options(self, parser):
        """
        Add options to the parser
        """

    def check_options(self, options, args):
        return True

    def __call__(self, prog_name, args):
        """
        Allows a BaseCommand to be a callable
        """
        self.run(prog_name, args)

    def run(self, prog_name, args):
        """
        Instantiates the parser checks the arguments. Afterwards is executes the
        command.
        """
        parser = self.get_option_parser(prog_name)
        options, rest = parser.parse_args(args)
        if not self.check_options(options, rest):
            return 2

        self.execute(options, rest)
        return 0

    def execute(self, options, rest):
        """
        Execute your command here
        """
