# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
# Torsten Irlaender <torsten.irlaender@intevation.de>
#

import os
import time
import re
import random
import hashlib
import logging
from codecs import getdecoder
from types import ListType, TupleType
import uuid

import psycopg2 as dbapi

from decorator import decorator

from webob.exc import HTTPUnauthorized, HTTPForbidden, HTTPNotFound

from pylons import request, config, app_globals as g

from mpulsweb.lib.db import DB, db, enter, leave
from mpulsweb.lib.translation import _

# normally we would import session from mpulsweb.lib.base, but to avoid
# circular imports we import it from app_globals.
from mpulsweb.lib.app_globals import session


log = logging.getLogger(__name__)


check_re = re.compile(r"^[a-z_0-9]+$")

FETCH_USER_DATA_SQL = """\
SELECT id, vorname, nachname, passwort_aendern, aktiviert,
       rolle, gid, last_login, vertreter
FROM ka_benutzer_tbl_view
WHERE login = %(login)s
"""
UPDATE_LASTLOGIN_SQL = """SELECT set_last_logintime(%(user_id)s)"""

ASCII_DECODER = getdecoder("ascii")

log = logging.getLogger(__name__)


def slashSplit(s):
    out = []
    for p in s.split("/"):
        if p.find("=") == -1 and out:
            out[-1] += '/' + p
        else:
            out.append(p)
    return out


def validPassword(s):
    try:
        ASCII_DECODER(s)
    except (UnicodeEncodeError, UnicodeDecodeError):
        return False
    return True


def checkLogin(user, password):
    """Check username and password.

    The user parameter should be the username the user typed into the
    login form. This function expands it to the full username used for
    the database using get_db_parameters and tries to open a database
    connection with this full login name and the password.

    The return value is a tuple consisting of the user object and the
    full login name used. One or both of these may be None. There are
    three cases:

       (None, None)

           The user name, database name or password do not pass a simple
           syntact check: the database name and username must consist of
           alpha-numerical ASCII characters and the password must be
           ASCII only.

       (None, full_login)

           The database connection could not be established. This
           usually means that the username and or password are
           incorrect, but it could also be a problem with the database.
           The reasons for the failure are logged.

       (userobject, full_login)

           The database connection could be established and therefore
           the user is considered to be logged in.

    In all cases wher the full login could be determined by this
    function, it is returned so that it can be used to record login
    failures and successes in order to show them to the user when the
    login successfully the next time.
    """
    login = None
    dbname = get_unmapped_db_name()
    if (not password or not validPassword(password)
            or not dbname or not check_re.match(dbname)
            or not check_re.match(user)):
        return (None, login)

    # Imported here because mpulsweb.model.user imports
    # mpulsweb.lib.security, which leads to circular imports.
    from mpulsweb.model.user import SessionUser, SessionSuperAdmin

    try:
        db_parameters = get_db_parameters(user, password,
                                          config.get('mpuls.db.host'),
                                          config.get('mpuls.db.port'),
                                          config.get('mpuls.db.database'),
                                          config.get('mpuls.db.user'))
        login = db_parameters["user"]
        mappeddb = db_parameters["database"]
        log.info('Login: user "%s" in "%s" -> "%s" in "%s"' %
                 (user, dbname, login, mappeddb))
        mydb = DB(**db_parameters)

        conn = None
        try:
            conn = mydb.getConnection()

            if user != 'adm':
                data = fetch_user_data(conn, login)
                userobject = SessionUser(login, password)
                userobject.setData(data)
                update_last_login_time(conn, data)
            else:
                userobject = SessionSuperAdmin(login, password)

            # return userobject
            log.info('Login ok :)')
            return (userobject, login)
        finally:
            mydb.closeConnections()
    except dbapi.OperationalError, err:
        # This is likely a login failure
        # XXX: Is there a way in DB Api 2.0 to determine
        # this database independent?
        log.error("OperationalError from database"
                  " while checking user credentials")
    except:
        log.exception("Exception while checking user credentials")
    log.info('Login failed :(')
    return (None, login)


def fetch_user_data(conn, login):
    data = None
    cursor = conn.cursor()
    fields = {'login': login}
    cursor.execute(FETCH_USER_DATA_SQL, fields)
    result = cursor.fetchone()
    data = {}
    data['id'] = result[0]
    data['firstname'] = result[1]
    data['lastname'] = result[2]
    data['newpass'] = result[3]
    data['activated'] = result[4]
    data['role'] = result[5]
    data['gid'] = result[6]
    data['last_login'] = result[7]
    data['standin'] = result[8]
    return data


def update_last_login_time(conn, data):
    cursor = conn.cursor()
    cursor.execute(UPDATE_LASTLOGIN_SQL,
                   {'user_id': data.get('id')})
    conn.commit()


def generateID(obj=None):
    array = [time.time(), os.times(), random.random()]
    if not obj is None:
        array.append(id(obj))
    return hashlib.md5(str(array)).hexdigest()


def subnet(addr):
    """poor man's subnet splitting"""
    x = addr.rsplit('.', 1)  # IPv4
    if len(x) == 2:
        return x[0]
    x = addr.rsplit(':', 1)  # IPv6
    if len(x) == 2:
        return x[0]
    return addr


def userIdentity():
    env = request.environ
    return "%s" % env.get('HTTP_USER_AGENT', '')


def no_csrf_token_required(function):
    """Decorator for POST action methods that do not need CSRF protection.

    This decorator simply adds the attribute csrf_token_required with a
    value of False to the function. The check_csrf_token function uses
    it to determine whether the CSRF token check does have to be
    performed for a given action.

    Use this sparingly! It's probably only useful in cases where the
    POST action behaves more like GET in that it does not change
    anything on the server and where it's expected that users will reuse
    the same form multiple times without using the browser's history. In
    mpulsweb this is only the case for evaluations.
    """
    function.csrf_token_required = False
    return function


def current_csrf_token():
    """Return the CSRF-token for the current request, creating it if necessary.
    The return value is a string with the hexadecimal representation of
    the token to use in all HTML-forms generated by the current request.
    It is the same for all calls for the same request.

    If there is no token associated with the current request, this
    function generates a new one, adds it to the request and stores it
    in the session.
    """
    token = getattr(request, "mpuls_csrf_token", None)
    if token is None:
        token = uuid.uuid4().hex
        request.mpuls_csrf_token = token
        session["CSRF_TOKEN"] = token
        session.save()
    return token


def check_csrf_token(request):
    """Check the CSRF protection for the request.
    The return value is True iff any of these conditions is met:

     - The request is not a POST request

     - The request is a POST request and the method implementing the
       action has a csrf_token_required attribute with a false value
       (see the no_csrf_token_required decorator)

     - The request is a POST request and has a parameter called
       'csrftoken' with a value that matches the corresponding value
       from the session.

    As a side-effect, for all POST requests the CSRF-token is deleted
    from the session, since it must be used at most once. Also, the
    'csrftoken' entry is removed from the request's POST dictionary to
    avoid problems with controllers that don't expect it, like the
    FormularpageController which passes the parameters to libformed
    which in turn expects all parameters to name existing fields.
    """
    if request.method != "POST":
        return True

    action = request.environ["pylons.routes_dict"]["action"]
    func = getattr(request.environ["pylons.controller"],
                   action.replace('-', '_'), None)
    if not getattr(func, "csrf_token_required", True):
        return True

    local_token = session.get("CSRF_TOKEN")
    if local_token is None:
        return False

    del session["CSRF_TOKEN"]
    session.save()

    request_token = request.POST.pop("csrftoken", None)
    return (request_token is not None
            and constant_time_equal(local_token, request_token))


def constant_time_equal(s1, s2):
    """Return whether the strings s1 and s2 are equal taking constant time.

    'Contant time' is meant in the sense that the time taken does not
    depend on which character is the first unequal character. If both
    strings differ in length, the function returns False immediately.
    """
    # This implementation is basically identical to that used by Django.
    if len(s1) != len(s2):
        return False

    difference = 0
    for (a, b) in zip(s1, s2):
        difference |= ord(a) ^ ord(b)
    return difference == 0


def get_dbname_from_http_header():
    """Return the KA-specific part of the DB-name set in HTTP-Header fields.

    This function expects the CN part of the Subject-DN from the
    SSL-client-certificate in either the MPULS_DBNAME or the
    SSL_CLIENT_S_DN_CN HTTP request header field.  The CN must consist
    of at least four whitespace separated words.  The KA-specific part
    of the database name is the fourth of those words.

    If the value cannot be determined, which usually means that neither
    of the header fields is set, the function returns None.
    """
    dbname = request.environ.get('MPULS_DBNAME')
    if not dbname:
        # MPULS_DBNAME not set or empty, fall back to older request
        # header
        log.debug("No database name defined in http-header MPULS_DBNAME,"
                  " falling back to SSL_CLIENT_S_DN_CN")
        dbname = request.environ.get('SSL_CLIENT_S_DN_CN')

    if not dbname:
        return None

    return dbname.split(' ')[3].lower()


def get_dbname_from_config():
    """Return the KA-specific part of the database name from the configuration.
    The return value is the value of the config entry 'mpuls.db.name'.
    """
    return config.get('mpuls.db.name')


def get_dbname_from_session():
    """Return the KA-specific part of the database name from the session.
    If it's not available in the session, the function returns None. The
    value should be stored in the session with store_dbname_in_session.
    """
    return session.get("MPULS_UNMAPPED_DBAGENCY")


def store_dbname_in_session(dbname):
    """Store the KA-specific part of the database name in the session.
    The value can be retrieved again with get_dbname_from_session. It is
    intended to be used if the name has to be determined by the
    application using a method not covered by get_unmapped_db_name. The
    application should call store_dbname_in_session before calling
    checkLogin.
    """
    session["MPULS_UNMAPPED_DBAGENCY"] = dbname
    session.save()


def get_unmapped_db_name():
    """Return the KA-specific part of the database name. In first place it
    tries to get the name from a saved db name in a session var.  If this ther
    is no value the value is taken from HTTP-Headers containing information
    from the SSL-client-certificate using get_dbname_from_http_header().  If
    those headers are not set, the value is taken from the configuration using
    get_dbname_from_config().
    """
    dbname = get_dbname_from_session()
    if not dbname:
        log.debug("Could not determine database name from Request params,"
                  " falling back to HTTP header ")
        dbname = get_dbname_from_http_header()

    if not dbname:
        log.debug("Could not determine database name from HTTP header,"
                  " falling back to config file")
        dbname = get_dbname_from_config()

    if not dbname:
        log.debug("Could not determine database name from configuration file")
        raise RuntimeError("Could not determine database name")

    return dbname


def getDbName():
    """Like get_unmapped_db_name, but apply the dbmapping to the name"""
    dbname = get_unmapped_db_name()
    mapped_parameters = g.mpuls_dbmapping.get_mapping().get(dbname)
    if mapped_parameters is not None:
        unmapped_dbname = dbname
        dbname = mapped_parameters.get("dbagency", dbname)
        log.debug("Mapping DB name '%s' on '%s'" % (unmapped_dbname, dbname))
    return dbname


def get_kaname_from_http_header():
    """Return the KA-name given by HTTP-Header fields.

    This function expects the Subject-DN from the SSL-client-certificate
    in either the MPULS_APPNAME or the SSL_CLIENT_S_DN HTTP request
    header field.  The return value is the value of the last OU
    component of that DN.  Subject DNs used for MPuls installations
    usually contain multuple OU fields with the last one describing the
    institution the user works for.

    If the value cannot be determined, which usually means that neither
    of the header fields is set, the function returns None.
    """

    subject_dn = request.environ.get('MPULS_APPNAME')
    if not subject_dn:
        # MPULS_APPNAME not set or empty, fall back to older request
        # header
        log.debug("HTTP-Header field MPULS_APPNAME not set or empty,"
                  " falling back to SSL_CLIENT_S_DN")
        subject_dn = request.environ.get('SSL_CLIENT_S_DN')
    if not subject_dn:
        return None

    for f in slashSplit(subject_dn)[::-1]:
        field = f.split('=')
        if field[0] == "OU":
            return field[1]

    log.debug("Subject DN found, but it did not contain an OU field")
    return None


def get_kaname_from_config():
    """Return the KA-name from the configuration"""
    return config.get('mpuls.app.name')


def getKAName():
    kaname = get_kaname_from_http_header()

    if not kaname:
        log.debug("Could not determine KA-name from HTTP header,"
                  " falling back to config file")
        kaname = get_kaname_from_config()

    if not kaname:
        log.debug("Could not determine KA-Name from configuration file")
        raise RuntimeError("Could not determine KA-name name")

    return kaname


def get_db_parameters(login, password, host, port, dbtemplate, usertemplate):
    """Determine database connection parameters from user-specific information.

    Parameters:

      login, password -- Login name and corresponding password

      host, port -- Default host and portname of the database cluster

      dbtemplate -- Default template string used to derive the actual
                    database name from dbagency

      usertemplate -- Default template string used to derive the actual
                      database username to use when logging into the
                      database, derived from dbagency and login name.

    The dbagency parameter is used to lookup agency specific settings in
    the configuration (via g.mpuls_dbmapping.get_mapping()) which can
    override any of the parameters host, port, dbtemplate, usertemplate
    and dbagency.

    The return value is a dictionary with the keys host, port, database,
    user and password.  It can be used directly as a dictionary for the
    keyword parameter of a DB-API connect() call.
    """
    dbagency = get_unmapped_db_name()
    mapping = g.mpuls_dbmapping.get_mapping()
    mapped_parameters = mapping.get(dbagency)
    if mapped_parameters:
        log.debug('mapped_parameters: %s' % mapped_parameters)
        dbagency = mapped_parameters.get('dbagency', dbagency)
        host = mapped_parameters.get('host', host)
        port = mapped_parameters.get('port', port)
        dbtemplate = mapped_parameters.get('dbtemplate', dbtemplate)
        usertemplate = mapped_parameters.get('usertemplate', usertemplate)

    return dict(host=host, port=port, database=dbtemplate % dbagency,
                user=usertemplate % (dbagency, login),
                password=password)


def checkRole(role):
    if type(role) in (ListType, TupleType):
        _role = role
    else:
        _role = (role,)

    def validate(func, self, *args, **kwargs):
        if not hasRole(_role):
            log.error('Not authorised to look at the requested document. '
                      'Returns an HTTP Unauthorized response.')
            raise HTTPUnauthorized(detail=_('You are not authorised to '
                                            'look at the requested document.'))
        try:
            return func(self, *args, **kwargs)
        except dbapi.OperationalError, err:
            # FIXME: An OperationalError at this point does not
            # necessarily indicate a permission problem, so converting
            # into it into a HTTPUnauthorized exception is not
            # necessarily the right thing, even though a permission
            # problem is a likely cause.  At the very least, log the
            # original exception because it's also likely to indicate a
            # problem with the database setup and discarding the
            # information of the original exception would make debugging
            # much harder.
            log.exception("Converting DB-API OperationalError into"
                          " an HTTP Unauthorized response")
            raise HTTPUnauthorized(detail=_('You are not authorised to '
                                            'look at the requested document.'))

    return decorator(validate)


def hasRole(rolelist):
    try:
        user = session['USER_AUTHORIZED']
        user_role = user.getPermissionRole()
        return user_role in rolelist
    except KeyError:
        return False


def check_feature(sec, feature):
    '''Helper function which checks if the feature is enabled in mpuls_config.
    If not raise an 404 Not Found on default.'''
    if not g.mpuls_config.is_enabled(sec, feature):
        raise HTTPNotFound()
    return True


def check_state(case, role2states):
    '''Helper function which checks if the state of the case is in state list
    of the role. Return False if not in this list.'''
    state = case.state.getState()
    try:
        user = session['USER_AUTHORIZED']
        user_role = user.getPermissionRole()
        return state in role2states.get(user_role, [])
    except KeyError:
        return False

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
