#!/bin/bash

# Usage:
#   dumpnormalizedschema.sh FKZ [PORT] [PREFX] > schema-FKZ.txt
#
# Dumps part of the DB schema to stdout. The output is normalized by
# replacing strings in various ways to make the output for different
# databases comparable with diff.
#
# The optional PREFIX argument is can be used for e.g. JMD and mpuls_S
# where additional databases exist whose names are derived from FKZ with
# by prefixing FKZ with PREFIX (thus, this prefix must not be part of
# FKZ already)
#
# This is useful to e.g. compare a newly installed DB with an older DB
# that was upgraded to the same state as the newly installed DB.

FKZ=$1
PORT=${2:-5432}
PREFIX=${3:-}
DBNAME=ka_${PREFIX}${FKZ}_db


normalize() {
    sed -f <(normalize_sed_program)
}

normalize_sed_program() {
cat <<EOF
# Replace FKZ numbers in DB and user-names with dummy FKZ XXXX
# Reason: FKZ numbers usually differ between DBs whose schemas we want
# to compare.
# If a prefix has been given on the command line, some of these names
# contain that prefix and some don't. Therefore, the prefix is optional
# in this pattern.
s/ka_\(${PREFIX}\)\?${FKZ}_/ka_XXXX_/g

# Remove table sizes mentioned in \dt+ output.
# Reason: Different DBs usually have different amounts of data in the
# same tables.
s/[0-9]\+ \(bytes\|kB\|MB\) *|/ <size removed> |/

# Remove last_value lines from sequence dumps.
# Reason: This value depends on e.g. the number of insertions done.
s/^ last_value.*/<last_value line removed>/

# Remove is_called lines from sequence dumps.
# Reason: Its value depends on the state of the sequence.
s/^ is_called.*/<is_called line removed>/

# Remove log_cnt lines from sequence dumps.
# Reason: Its value depends on the state of the sequence.
s/^ log_cnt.*/<log_cnt line removed>/

# mpuls_S:
# At one point abschluss_tbl was renamed to abschluss_bund_tbl.
# Automatically generated sequences and constraints still use the old
# generated names in older databases, whereas in newer databases the
# sequences/constraints have names derived from the new table name.
# Rename them to the new names
#
# This rule is specific to the mpuls_S stipendium part, but it should
# not be a problem in other mpuls applications.
s/\<abschluss_tbl_id_seq\>/abschluss_bund_tbl_id_seq/
s/\<abschluss_tbl_pkey\>/abschluss_bund_tbl_pkey/
EOF
}


# Print the names of all tables that are used for the options of choice
# fields, sorted by table name. A relatively simply heuristic for
# determining which tables to list without having to hard-code
# application specific information is to simply list all tables which
# have exactly two columns which are named "id" and "value".
list_choice_tables() {
    psql -p $PORT -d "$DBNAME" -t -c "
        SELECT table_name FROM information_schema.columns
         WHERE table_schema = 'public'
           AND table_name LIKE '%_tbl'
      GROUP BY table_name
        HAVING array_agg((column_name :: text) ORDER BY column_name)
               IN (ARRAY['beschreibung','id'], ARRAY['id','value'])
      ORDER BY table_name;"
}


echo "Definitions of database ${DBNAME} on port ${PORT}"

(psql -p $PORT -d $DBNAME | normalize) <<EOF
\dt+
\dv+
\df+ public.*
\d+ public.*
EOF

echo "Contents of choice tables"

for table in $(list_choice_tables); do
    echo "$table"
    psql -p $PORT -d $DBNAME -c "SELECT * FROM $table ORDER BY id;"
done
