SET ROLE :adm_ka_owner;
BEGIN;

-- ka_benutzergruppe_tbl
-- Tabelle fuer die Benutzergruppen. Jeder Benutzer erhaelt eine
-- Benutzergruppe in der er exklusiv Mitglied ist.
-- Weitere Benutzergruppen können angelegt werden, die weitere Nutzer
-- enthalten können

CREATE TABLE ka_benutzergruppe_tbl
(
        id serial NOT NULL,
        name varchar(128) UNIQUE NOT NULL,
        CONSTRAINT ka_benutzergruppe_tbl_pkey PRIMARY KEY (id)
);

-- Benutzertabelle
-- Diese Tabelle enthält erweiterte Angaben zu den Nutzern der KA Datenbank.
-- Das Feld 'login' enthält hier den vollen Login-Namen, so wie er clusterweit
-- in der Datenbank bekannt ist (ka_<agentur_<loginname>)

CREATE TABLE ka_benutzer_tbl
(
        id serial NOT NULL,
        login varchar(64) NOT NULL,
        vorname varchar(64),
        nachname varchar(64),
        telefon varchar(32),
        raumnummer varchar(16),
        filiale varchar(64),
        aktiviert bool DEFAULT true,
        passwort_aendern bool DEFAULT true,
        last_login timestamp,
        gid       int4 REFERENCES ka_benutzergruppe_tbl(id) ON DELETE CASCADE,
        vertreter int4 REFERENCES ka_benutzergruppe_tbl(id) ON DELETE SET DEFAULT,
        CONSTRAINT ka_benutzer_tbl_pkey PRIMARY KEY (id)
);

GRANT SELECT ON ka_benutzer_tbl TO GROUP :cm_group;
GRANT SELECT ON ka_benutzer_tbl TO GROUP :admin_group;
GRANT UPDATE ON ka_benutzer_tbl TO GROUP :admin_group;

-- Benutzer-Fallakte Transitionstabelle
-- Diese Tabelle verbindet Nutzer mit Fallakten, und legt auf diese Weise
-- fest welche Benutzer ausser dem initialen Bearbeiter oder dem KA-Admin
-- Zugriff auf die Fallakte haben.

CREATE TABLE nm_benutzer_master_tbl (
    id serial PRIMARY KEY,
    benutzer_id int4,
    master_id int4
);

-- Benutzer Fallakten Transitionstabelle
ALTER TABLE nm_benutzer_master_tbl ADD CONSTRAINT has_group_fk FOREIGN KEY
(benutzer_id) REFERENCES ka_benutzergruppe_tbl (id) ON DELETE CASCADE;

GRANT SELECT ON ka_benutzer_tbl TO GROUP :cm_group;
GRANT SELECT ON ka_benutzer_tbl TO GROUP :admin_group;

-- ka_nm_benutzer_gruppe_tbl
-- Tabelle zur Zuordnung zwischen Benutzern und Benutzergruppen.

CREATE TABLE nm_benutzer_gruppe_tbl
(
        id serial NOT NULL,
        gruppe_id int4 REFERENCES ka_benutzergruppe_tbl(id) ON DELETE CASCADE,
        benutzer_id int4 REFERENCES ka_benutzer_tbl(id) ON DELETE CASCADE,
        CONSTRAINT nm_benutzer_gruppe_tbl_pkey PRIMARY KEY (id)
);

-- getuserid()
-- holt id des current_user aus der ka_benutzer_tbl
--
CREATE OR REPLACE FUNCTION getuserid() RETURNS integer
    LANGUAGE 'plpgsql' STABLE
    AS $$
DECLARE
myid integer;
BEGIN
SELECT INTO myid id FROM ka_benutzer_tbl WHERE login = session_user;
RETURN myid;
END;
$$;

-- Definition von DB-seitigen Funktionen zur Nutzerverwaltung
-- Autor: Frank Koormann <frank@intevation.de>
-- Stand: 2007-11-01

-- Anlegen einer Rolle in der DB einer KA:
-- - Ueber mpuls_create_role() Datenbank-Rolle anlegen und
-- - dann Eintragungen in ka_benutzer_tbl machen.
CREATE OR REPLACE FUNCTION ka_create_role (
        agency          varchar(52),
        grouprole       varchar(64),
        loginname       varchar(64),
        firstname       varchar(64),
        surname         varchar(64),
        phone           varchar(64),
        room            varchar(64),
        branchoffice    varchar(64),
        activated       boolean,
        setpassword     boolean
    ) RETURNS varchar(13)
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
    newrole    RECORD;
    groupid    integer;
    userid     integer;
BEGIN
    SELECT INTO newrole * FROM mpuls_create_role(loginname, grouprole, agency)
        AS (db_name varchar(64), db_user varchar(64), pw varchar(13));
    EXECUTE 'GRANT CONNECT ON DATABASE ' || newrole.db_name
                || ' TO ' || newrole.db_user;
    INSERT INTO ka_benutzer_tbl ( login, vorname, nachname,
        telefon, raumnummer, filiale,
        aktiviert, passwort_aendern, gid)
        VALUES
        (newrole.db_user, firstname, surname,
        phone, room, branchoffice,
        activated, setpassword, NULL);
    SELECT INTO userid currval('ka_benutzer_tbl_id_seq');
    SELECT INTO groupid create_benutzergruppe_ds(surname ||', '|| firstname ||' ('||loginname||')');
    UPDATE ka_benutzer_tbl SET gid = groupid WHERE id = userid;
    INSERT INTO nm_benutzer_gruppe_tbl (benutzer_id, gruppe_id) VALUES (userid, groupid);

    RETURN newrole.pw::varchar(13);
END;
$$ ;

REVOKE ALL ON FUNCTION ka_create_role(varchar, varchar, varchar,
    varchar, varchar, varchar, varchar, varchar, boolean, boolean)
        FROM PUBLIC;
GRANT EXECUTE ON FUNCTION ka_create_role(varchar, varchar, varchar,
    varchar, varchar, varchar, varchar, varchar, boolean, boolean)
        TO GROUP :admin_group;

CREATE OR REPLACE FUNCTION ka_delete_role (
        loginname       varchar(64),
        agency          varchar(52)
    ) RETURNS VOID
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
    db_user  varchar(128);
    groupid   integer;
BEGIN
    db_user := quote_ident('ka_' || agency || '_' || loginname);
    SELECT INTO groupid gid FROM ka_benutzer_tbl WHERE login = db_user;
    DELETE FROM ka_benutzer_tbl WHERE login = db_user;
    DELETE FROM ka_benutzergruppe_tbl WHERE id = groupid;
    PERFORM mpuls_drop_role(loginname, agency);
END;
$$ ;

REVOKE ALL ON FUNCTION ka_delete_role(varchar, varchar) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION ka_delete_role(varchar, varchar)
            TO GROUP :admin_group;

-- Funktionen zu Erstellen und Loeschen neuer Eintraeger in der
-- Benutzergruppen Tabelle
CREATE OR REPLACE FUNCTION create_benutzergruppe_ds(gname varchar) RETURNS integer
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    INSERT INTO ka_benutzergruppe_tbl (name) VALUES (gname);
    RETURN currval('ka_benutzergruppe_tbl_id_seq');
END;
$$ ;

REVOKE ALL ON FUNCTION create_benutzergruppe_ds( varchar ) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION create_benutzergruppe_ds( varchar ) TO GROUP :admin_group;

CREATE OR REPLACE FUNCTION delete_benutzergruppe_ds(gid integer) RETURNS void
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
BEGIN
    DELETE FROM ka_benutzergruppe_tbl WHERE id = gid;
END;
$$ ;

REVOKE ALL ON FUNCTION delete_benutzergruppe_ds( integer ) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION delete_benutzergruppe_ds( integer ) TO GROUP :admin_group;

--
-- Voreinstellung Vertretungsregeln setzen
CREATE OR REPLACE FUNCTION ka_set_standin (standin_id integer, user_id integer)
    RETURNS VOID
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    UPDATE ka_benutzer_tbl_view set vertreter = standin_id WHERE id = user_id;
END;
$$ ;

REVOKE ALL ON FUNCTION ka_set_standin(integer, integer) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION ka_set_standin ( integer, integer ) TO :cm_group;

-- Funktion um die Flagge zurueckzusetzten, dass der Benutzer sein Passwort bei
-- der Neuanmeldung neu setzen muss. Die Funktion muss auch fuer die Benutzer
-- zugaenglich sein.
CREATE OR REPLACE FUNCTION ka_unset_passrefresh_flag ( userlogin varchar )
    RETURNS VOID
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    UPDATE ka_benutzer_tbl_view set passwort_aendern = False
            WHERE login = userlogin;
END;
$$ ;

REVOKE ALL ON FUNCTION ka_unset_passrefresh_flag(varchar) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION ka_unset_passrefresh_flag ( userlogin varchar )
                TO :admin_group;
GRANT EXECUTE ON FUNCTION ka_unset_passrefresh_flag ( userlogin varchar )
                TO :cm_group;

-- Funktion zum Setzen des Benutzerpassworts
CREATE OR REPLACE FUNCTION ka_set_password(loginname varchar, pw varchar)
    RETURNS text
    LANGUAGE 'plpgsql'
    AS $$
BEGIN
    EXECUTE 'ALTER USER ' || loginname || ' WITH PASSWORD '
              || quote_literal(pw);
    EXECUTE ka_unset_passrefresh_flag(loginname);
          RETURN pw::text;
    RETURN ''::text;
END;
$$ ;

REVOKE ALL ON FUNCTION ka_set_password(varchar, varchar) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION ka_set_password(varchar, varchar)
                TO :admin_group;
GRANT EXECUTE ON FUNCTION ka_set_password(varchar, varchar)
                TO :cm_group;

-- Funktionen zum Hinzufuegen und Loeschen der Benutzer in die
-- Benutzer-Fallakten Transitionstabelle, sowie zum Setzen des Bearbeiters
-- einer Fallakte.
CREATE OR REPLACE FUNCTION create_nm_benutzer_master_ds ( userid integer,
    masterid integer ) RETURNS integer
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    INSERT INTO nm_benutzer_master_tbl (benutzer_id, master_id)
        VALUES (userid, masterid);
    RETURN currval('nm_benutzer_master_tbl_id_seq');
END;
$$ ;

REVOKE ALL ON FUNCTION create_nm_benutzer_master_ds(integer, integer)
                FROM PUBLIC;
GRANT EXECUTE ON FUNCTION create_nm_benutzer_master_ds(integer, integer)
                TO GROUP :cm_group;
GRANT EXECUTE ON FUNCTION create_nm_benutzer_master_ds(integer, integer)
                TO GROUP :admin_group;


CREATE OR REPLACE FUNCTION delete_nm_benutzer_master_ds ( userid integer,
    masterid integer ) RETURNS VOID
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    DELETE FROM nm_benutzer_master_tbl
            WHERE master_id = masterid AND benutzer_id = userid;
END;
$$ ;

REVOKE ALL ON FUNCTION delete_nm_benutzer_master_ds(integer, integer)
                FROM PUBLIC;
GRANT EXECUTE ON FUNCTION delete_nm_benutzer_master_ds(integer, integer)
                TO GROUP :cm_group;
GRANT EXECUTE ON FUNCTION delete_nm_benutzer_master_ds(integer, integer)
                TO GROUP :admin_group;


-- Funktionen zu Erstellen und Loeschen neuer Eintraeger in der
-- nm_benutzer_gruppe Tabelle
CREATE OR REPLACE FUNCTION create_nm_benutzer_gruppe_ds(bid integer, gid integer) RETURNS integer
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    INSERT INTO nm_benutzer_gruppe_tbl (benutzer_id, gruppe_id) VALUES (bid, gid);
    RETURN currval('nm_benutzer_gruppe_tbl_id_seq');
END;
$$ ;
REVOKE ALL ON FUNCTION create_nm_benutzer_gruppe_ds( integer, integer ) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION create_nm_benutzer_gruppe_ds( integer, integer ) TO GROUP :admin_group;

CREATE OR REPLACE FUNCTION delete_nm_benutzer_gruppe_ds( bid integer, gid integer) RETURNS void
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    DELETE FROM nm_benutzer_gruppe_tbl WHERE gruppe_id = gid AND benutzer_id = bid;
END;
$$ ;
REVOKE ALL ON FUNCTION delete_nm_benutzer_gruppe_ds( integer, integer ) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION delete_nm_benutzer_gruppe_ds( integer, integer ) TO GROUP :admin_group;

-- Funktionen zum Setzten des Editors
CREATE OR REPLACE FUNCTION set_case_editor (userid integer, masterid integer)
    RETURNS VOID
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    UPDATE master_tbl set bearbeiter_id = userid WHERE id = masterid;
END;
$$ ;

REVOKE ALL ON FUNCTION set_case_editor(integer, integer) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION set_case_editor(integer, integer)
                TO GROUP :admin_group;

CREATE OR REPLACE FUNCTION set_case_editor_name (firstname varchar, lastname varchar, masterid integer)
    RETURNS VOID
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    UPDATE master_tbl set mitarb_zv = firstname, mitarb_zn = lastname WHERE id = masterid;
END;
$$ ;

REVOKE ALL ON FUNCTION set_case_editor_name(varchar, varchar, integer) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION set_case_editor_name(varchar, varchar, integer)
                TO GROUP :admin_group;

CREATE OR REPLACE FUNCTION getrole( user_id integer ) RETURNS varchar
    LANGUAGE 'plpgsql' STABLE SECURITY DEFINER
    AS $$
DECLARE
    role varchar;
    oid integer;
BEGIN
    SELECT p.usesysid INTO oid FROM pg_user p JOIN ka_benutzer_tbl b
        ON b.login = p.usename
        WHERE b.id = user_id;
    SELECT groname INTO role FROM pg_group WHERE oid = ANY (grolist);
    RETURN role;
END;
$$;

REVOKE ALL ON FUNCTION getrole(user_id integer) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION getrole( integer ) TO GROUP :cm_group;
GRANT EXECUTE ON FUNCTION getrole( integer ) TO GROUP :admin_group;

--
-- getuserid()
-- holt id des current_user aus der ka_benutzer_tbl
--
-- Funktion wurde in die 00precond.sql verschoben

--
-- getoid()
-- OID des current_user
--
CREATE OR REPLACE FUNCTION getoid() RETURNS integer
    LANGUAGE 'plpgsql' STABLE
    AS $$
DECLARE
    myoid OID;
BEGIN
    SELECT INTO myoid oid FROM pg_roles WHERE rolname = current_user;
    RETURN myoid;
END;
$$;

REVOKE ALL ON FUNCTION getoid() FROM PUBLIC;
GRANT EXECUTE ON FUNCTION getoid() TO GROUP :cm_group;
GRANT EXECUTE ON FUNCTION getoid() TO GROUP :admin_group;

-- isEditorOrStandin(integer)
-- Liefert TRUE wenn der Benutzer fuer die Fallakte
-- a) Bearbeiter oder
-- b) Vertreter ist
-- Sonst FALSE
--

CREATE OR REPLACE FUNCTION isEditorOrStandin(mid integer)
    RETURNS boolean
    LANGUAGE 'plpgsql' STABLE
    AS $$
DECLARE
    bid integer;
    isAllowed integer;
BEGIN
    SELECT bearbeiter_id into bid from master_tbl_view WHERE id = mid;
    IF bid = getuserid() THEN
        RETURN TRUE;
    ELSE
        SELECT INTO isAllowed benutzer_id FROM
        nm_benutzer_master_tbl_view
        WHERE master_id = mid
        AND benutzer_id IN (
                SELECT gruppe_id FROM nm_benutzer_gruppe_tbl_view
                WHERE benutzer_id = getuserid()
                );
    END IF;
    RETURN FOUND;
END;
$$;

REVOKE ALL ON FUNCTION isEditorOrstandin(integer) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION isEditorOrStandin(integer) TO GROUP :cm_group;
GRANT EXECUTE ON FUNCTION isEditorOrStandin(integer) TO GROUP :admin_group;

COMMIT;
-- Trigger/Logging
CREATE TRIGGER t_ka_benutzer_tbl_insert AFTER
    INSERT ON ka_benutzer_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE TRIGGER t_ka_benutzer_tbl_update AFTER
    UPDATE ON ka_benutzer_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE TRIGGER t_ka_benutzer_tbl_delete AFTER
    DELETE ON ka_benutzer_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE TRIGGER t1_nm_benutzer_master_tbl_insert AFTER
    INSERT ON nm_benutzer_master_tbl
    FOR EACH ROW EXECUTE PROCEDURE rg_lastaccess_log();

CREATE TRIGGER t1_nm_benutzer_master_tbl_update AFTER
    UPDATE ON nm_benutzer_master_tbl
    FOR EACH ROW EXECUTE PROCEDURE rg_lastaccess_log();

CREATE TRIGGER t1_nm_benutzer_master_tbl_delete AFTER
    DELETE ON nm_benutzer_master_tbl
    FOR EACH ROW EXECUTE PROCEDURE rg_lastaccess_log();
-- Create Views.
\i sql/install/views/01_users.sql

