SET ROLE :adm_ka_owner;
BEGIN;

CREATE TABLE ka_logbuch_art_tbl (
    id         SERIAL NOT NULL,
    beschreibung VARCHAR(256),
    CONSTRAINT ka_logbuch_art_tbl_pkey PRIMARY KEY (id)
);

GRANT SELECT ON ka_logbuch_art_tbl TO GROUP :cm_group;
GRANT SELECT ON ka_logbuch_art_tbl TO GROUP :admin_group;

CREATE TABLE ka_logbuch_typ_tbl (
    id         SERIAL NOT NULL,
    beschreibung VARCHAR(256),
    CONSTRAINT ka_logbuch_typ_tbl_pkey PRIMARY KEY (id)
);

GRANT SELECT ON ka_logbuch_typ_tbl TO GROUP :cm_group;
GRANT SELECT ON ka_logbuch_typ_tbl TO GROUP :admin_group;

CREATE TABLE ka_logbuch_tbl (
    id         SERIAL NOT NULL,
    master_id  int4 REFERENCES master_tbl(id)         ON DELETE CASCADE,
    uuid_id    CHAR(36),
    datum      TIMESTAMP,
    bearbeiter VARCHAR(256),
    art        int4 REFERENCES ka_logbuch_art_tbl(id) ON DELETE SET DEFAULT,
    typ        int4 REFERENCES ka_logbuch_typ_tbl(id) ON DELETE SET DEFAULT,
    kurz_notiz VARCHAR(256),
    notiz      text,
    dauer      interval,
    UNIQUE (uuid_id),
    CONSTRAINT ka_logbuch_tbl_pkey PRIMARY KEY (id)
);


-- Convert plain-text to HTML for the notiz column
--
-- The notiz-column contains HTML so when specifying the text for notiz
-- as plain-text, this function must be used to convert correctly.
CREATE OR REPLACE FUNCTION notiz_from_plaintext(notiz text) RETURNS text
AS $$
BEGIN
    RETURN  '<p>' || replace(replace(replace(replace(
                                notiz, '&', '&amp;'),
                                       '<', '&lt;'),
                                       '>', '&gt;'),
                                       E'\n', '<br>') || '</p>';
END;
$$ LANGUAGE plpgsql;


-- Insert application specific entries into the tables

-- Funktionen zum Anlegen und Loeschen von Logbucheintraegen
--
CREATE OR REPLACE FUNCTION create_logbuch_ds(mid integer, nuuid CHAR(36)) RETURNS integer
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
AS $$
DECLARE
    lid integer;
    editor_name varchar;
BEGIN
    IF NOT (isEditorOrStandin(mid) OR pg_has_role('admin','MEMBER')) THEN
        RAISE EXCEPTION 'DS ID not session_user ID';
    END IF;

    SELECT INTO editor_name nachname || ', ' || vorname from ka_benutzer_tbl WHERE id = getuserid();
    INSERT INTO ka_logbuch_tbl(master_id, datum, bearbeiter) VALUES (mid, now(), editor_name);

    SELECT INTO lid currval('ka_logbuch_tbl_id_seq');

    IF nuuid IS NOT NULL THEN
        UPDATE ka_logbuch_tbl SET uuid_id = nuuid WHERE id = lid;
    ELSE
        PERFORM set_uuid('ka_logbuch_tbl', lid);
    END IF;

    RETURN lid;
END;
$$;

REVOKE ALL    ON FUNCTION create_logbuch_ds(integer, CHAR(36)) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION create_logbuch_ds(integer, CHAR(36)) TO GROUP :cm_group;
GRANT EXECUTE ON FUNCTION create_logbuch_ds(integer, CHAR(36)) TO GROUP :admin_group;

CREATE OR REPLACE FUNCTION delete_logbuch_ds(tid integer) RETURNS void
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
    mid integer;
BEGIN
    SELECT INTO mid master_id FROM ka_logbuch_tbl WHERE id = tid;
    IF NOT isEditorOrStandin(mid) THEN
        RAISE EXCEPTION 'DS ID not session_user ID';
    END IF;
    DELETE FROM ka_logbuch_tbl WHERE id = tid;
END;
$$ ;

REVOKE ALL    ON FUNCTION delete_logbuch_ds(integer) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION delete_logbuch_ds(integer) TO GROUP :cm_group;


CREATE OR REPLACE FUNCTION delete_all_logbuch(mid integer)
RETURNS void
LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
AS $$
BEGIN
    DELETE FROM ka_logbuch_tbl WHERE master_id = mid;
END;
$$;

REVOKE ALL     ON FUNCTION delete_all_logbuch(integer) FROM PUBLIC;
GRANT  EXECUTE ON FUNCTION delete_all_logbuch(integer) TO GROUP :admin_group;

CREATE OR REPLACE FUNCTION anonymize_logbuch(mid integer)
RETURNS void
LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
AS $$
BEGIN
    DELETE FROM ka_logbuch_dokumente_tbl
    WHERE logbuch_id IN (SELECT id FROM ka_logbuch_tbl WHERE master_id = mid);

    -- DELETE FROM ka_logbuch_tbl WHERE master_id = mid;
    UPDATE ka_logbuch_tbl SET
        uuid_id    = NULL,
        bearbeiter = md5(lower(trim(both ' ' from coalesce(bearbeiter, '')))),
        kurz_notiz = NULL,
        notiz      = NULL
    WHERE master_id = mid;
END;
$$;

REVOKE ALL     ON FUNCTION anonymize_logbuch(integer) FROM PUBLIC;
GRANT  EXECUTE ON FUNCTION anonymize_logbuch(integer) TO GROUP :admin_group;

-- Trigger/Logging
CREATE TRIGGER t_ka_logbuch_tbl_insert AFTER
    INSERT ON ka_logbuch_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE TRIGGER t_ka_logbuch_tbl_update AFTER
    UPDATE ON ka_logbuch_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE TRIGGER t_ka_logbuch_tbl_delete AFTER
    DELETE ON ka_logbuch_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE TRIGGER t1_ka_logbuch_tbl_insert AFTER
    INSERT ON ka_logbuch_tbl
    FOR EACH ROW EXECUTE PROCEDURE rg_lastaccess_log();

CREATE TRIGGER t1_ka_logbuch_tbl_update AFTER
    UPDATE ON ka_logbuch_tbl
    FOR EACH ROW EXECUTE PROCEDURE rg_lastaccess_log();

CREATE TRIGGER t1_ka_logbuch_tbl_delete AFTER
    DELETE ON ka_logbuch_tbl
    FOR EACH ROW EXECUTE PROCEDURE rg_lastaccess_log();

-- Trigger fuer das automatische Anlegen eines Logbucheintrages
-- beim Anlegen eines Masterdatensatzes.
CREATE OR REPLACE FUNCTION master_ds_created_logbook() RETURNS trigger AS $master_ds_created_logbook$
DECLARE
    lid integer;
BEGIN
    SELECT INTO lid create_logbuch_ds(NEW.id, NULL);
    UPDATE ka_logbuch_tbl_view
        SET art = 0, typ = -1,
            notiz = notiz_from_plaintext('Fallakte angelegt.')
    WHERE
        id = lid;
    RETURN NEW;
END;
$master_ds_created_logbook$ LANGUAGE plpgsql;

CREATE TRIGGER master_ds_created_logbook_trigger AFTER
    INSERT ON master_tbl
    FOR EACH ROW EXECUTE PROCEDURE master_ds_created_logbook();

-- Example
--

-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (-2, 'Sonstige Angaben');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (-1, 'Keine Angabe');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (0, 'Automatischer Eintrag');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (1, 'Gespräch mit dem Jugendlichen');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (2, 'Kompetenzfeststellung');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (3, 'Korrespondenz mit dem/der Jugendlichen (Telefonate/Schriftverkehr)');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (4, 'Begleitung des/der Jugendlichen zu externen Stellen');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (5, 'Sonstige Betreuung des Jugendlichen');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (6, 'Gespräch mit Eltern/Personensorgeberechtigten');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (7, 'Korrespondenz mit Eltern/Personensorgeberechtigten (Telefonat/Schriftverkehr)');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (8, 'Sonstige Maßnahmen der Elternarbeit');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (9, 'Schule');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (10, 'Jugendhilfe');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (11, 'beteiligte Dritten/Kooperationspartnern');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (12, 'Förder bzw. Fallkonferenz');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (13, 'andere');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (14, 'Fahrtwege');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (15, 'Dokumentation WASKU');
-- INSERT INTO ka_logbuch_art_tbl (id, beschreibung) VALUES  (16, 'Teambesprechung');

-- INSERT INTO ka_logbuch_typ_tbl (id, beschreibung) VALUES (0, 'nicht aufsuchend');
-- INSERT INTO ka_logbuch_typ_tbl (id, beschreibung) VALUES (1, 'aufsuchend');
-- INSERT INTO ka_logbuch_typ_tbl (id, beschreibung) VALUES (-1, 'Keine Angabe');



--
-- Documents associated with logbook entries
--
-- These documents are not stored in ka_dokumente_tbl like the other
-- documents, but in a new table, ka_logbuch_dokumente_tbl. This appears
-- to be the simpler approach because the case distinctions that would
-- have to be made with rows of ka_dokumente_tbl would become even more
-- complicated.
--
-- Also a separate table avoids redundancy in the data model because
-- logbook entries are associated with a case and hence reference
-- master_tbl and ka_dokumente_tbl rows can also do that. That either
-- means duplication for logbook documents -- master_id stored twice, in
-- ka_dokumente_tbl and the corresponding ka_logbuch_tbl row -- or
-- master_id would be empty and not empty in
-- ka_logbuch_dokumente_tbl_view because it's convenient to have the
-- master_id available in the view.

CREATE TABLE ka_logbuch_dokumente_tbl
(
        id serial NOT NULL,
        uuid_id CHAR(36) UNIQUE,
        logbuch_id int4 REFERENCES ka_logbuch_tbl(id) ON DELETE CASCADE,
        groesse int4,
        mime varchar(256),
        daten bytea,
        name varchar(256),
        datum date,
        CONSTRAINT ka_logbuch_dokumente_tbl_pkey PRIMARY KEY (id)
);

-- Funktionen, zum Anlegen und Löschen von Dokumenten
CREATE OR REPLACE FUNCTION create_logbuch_dokument_ds(lid integer, nuuid char(36), filename varchar) RETURNS integer
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
        mid integer;
        did integer;
        rid integer;
BEGIN
    SELECT INTO mid master_id FROM ka_logbuch_tbl WHERE id = lid;
    IF NOT isEditorOrStandin(mid) THEN
        RAISE EXCEPTION 'DS ID not session_user ID';
    END IF;

    SELECT INTO did id from ka_logbuch_dokumente_tbl
     WHERE name = filename AND lid = logbuch_id;
    IF NOT FOUND THEN
        INSERT INTO ka_logbuch_dokumente_tbl(logbuch_id)
             VALUES (lid);
        SELECT INTO rid currval('ka_logbuch_dokumente_tbl_id_seq');
    ELSE
        RETURN did;
    END IF;

    IF nuuid IS NOT NULL THEN
        UPDATE ka_logbuch_dokumente_tbl
        SET    uuid_id = nuuid
        WHERE  id = rid;
    ELSE
        PERFORM set_uuid('ka_logbuch_dokumente_tbl', rid);
    END IF;
    RETURN rid;
END;
$$ ;

REVOKE ALL ON FUNCTION create_logbuch_dokument_ds(integer, char(36), varchar) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION create_logbuch_dokument_ds( integer, char(36), varchar) TO GROUP :cm_group;

CREATE OR REPLACE FUNCTION delete_logbuch_dokument_ds(did integer) RETURNS void
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
    mid integer;
BEGIN
    SELECT INTO mid t.master_id
      FROM ka_logbuch_dokumente_tbl d
      JOIN ka_logbuch_tbl t ON d.logbuch_id = t.id
     WHERE d.id = did;
    IF NOT isEditorOrStandin(mid) THEN
        RAISE EXCEPTION 'DS ID not session_user ID';
    END IF;
    DELETE FROM ka_logbuch_dokumente_tbl WHERE id = did;
END;
$$ ;

REVOKE ALL ON FUNCTION delete_logbuch_dokument_ds(integer) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION delete_logbuch_dokument_ds(integer)
                TO GROUP :cm_group;


-- Trigger/Logging
CREATE TRIGGER t_ka_logbuch_dokumente_tbl_insert AFTER
    INSERT ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE TRIGGER t_ka_logbuch_dokumente_tbl_update AFTER
    UPDATE ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();

CREATE TRIGGER t_ka_logbuch_dokumente_tbl_delete AFTER
    DELETE ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE mpuls_log();


CREATE OR REPLACE FUNCTION logbuch_dokumente_lastaccess_log()
RETURNS trigger AS $$
DECLARE
    mid integer;
    lid integer;
BEGIN
IF TG_OP = 'DELETE' THEN
   lid = OLD.logbuch_id;
ELSE
   lid = NEW.logbuch_id;
END IF;

SELECT INTO mid master_id
  FROM ka_logbuch_tbl_view
 WHERE id = lid;

PERFORM set_lastaccess_time(mid);

IF TG_OP = 'DELETE' THEN
   return OLD;
ELSE
   return NEW;
END IF;

END;
$$ LANGUAGE plpgsql;


CREATE TRIGGER t1_ka_logbuch_dokumente_tbl_insert AFTER
    INSERT ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE logbuch_dokumente_lastaccess_log();

CREATE TRIGGER t1_ka_logbuch_dokumente_tbl_update AFTER
    UPDATE ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE logbuch_dokumente_lastaccess_log();

CREATE TRIGGER t1_ka_logbuch_dokumente_tbl_delete AFTER
    DELETE ON ka_logbuch_dokumente_tbl
    FOR EACH ROW EXECUTE PROCEDURE logbuch_dokumente_lastaccess_log();

COMMIT;
-- Create Views.
\i sql/install/views/08_logbook.sql
