import logging

import formed.meta.data as data
from formed.meta.data import InfoLeaf, SaveLeaf
from formed.instance.tree import RepeatInstanceContainerNode, SetDataException
from formed.instance.backends.postgres import DBFactory
from formed.instance.backends.xmlimport import XMLImporter, XMLImportException


log = logging.getLogger(__name__)


class DBXMLImporter(XMLImporter):

    """XMLImporter that writes the imported cases into a database"""

    def import_xml(self, db, infile):
        """Read cases from the file object infile and update the database db.
        The db parameter must conform to the interface
        formed.instance.backends.database.DB.

        The import_case method is called once for each instance tree
        read from the xml file.  It is called with the instance tree
        factory, a database cursor and the structure instance tree.
        Derived classes may want to override that method.

        All database changes are done in a transaction.  If any exception is
        raised during the import, that transaction is rolled back, so no
        database changes will be made.
        """
        factory = DBFactory(self.structure_tree.meta_tree, db)
        db.begin()
        cursor = None
        try:
            try:
                cursor = db.acquire_cursor()
                for case in self.read_cases_from_xml(infile):
                    self.import_case(factory, cursor, case)
                db.commit()
            except:
                db.rollback()
                raise
        finally:
            db.release_cursor(cursor)

    def import_case(self, factory, cursor, components):
        """Write the case into the database using the given db cursor.
        The factory parameter is a DBFactory object that can be used to
        create new cases in the database.  The cursor is a cursor for
        the same database connection used by the factory so that all
        changes made through the cursor are part of the same transaction
        as the changes done via the factory.  The commit -- or rollback
        in the case of errors -- will be done by the caller of this
        method.

        The return value is the new or updated instance tree object.

        Derived classes may override or extend this method to
        e.g. determine whether to import a particular case at all.
        """
        return update_instance_in_database(components["master"], factory)


def update_instance_in_database(src, factory):
    """Update or create the database instance corresponding to src.

    The parameter src must be a structure instance tree.  If the case
    described by src already exists in the database as determined by
    their uuids, that case is updated.  Otherwise a new case is created
    in the database.

    The return value is the instance tree the case was stored into.
    """
    uuid = src.uuid
    dest_tree = None
    match_uuids = False
    if uuid:
        dest_tree = factory.loadInstanceTreeByUUID(uuid)
        match_uuids = True
    if dest_tree is None:
        dest_tree = factory.createNewInstanceTree(uuid)
    copy_to_full_instance_tree(src, dest_tree, match_uuids=match_uuids)
    return dest_tree


def copy_to_full_instance_tree(src, dest, match_uuids):
    data = dict()
    _copy_data_recursively(src.root, dest.getRootNode(), data,
                           match_uuids=match_uuids)
    try:
        dest.setData(data, convert=False, find_best_changes=False)
    except SetDataException, e:
        log.exception("SetDataException when loading a case from XML")
        raise XMLImportException("The case with uuid '%s' could not be imported"
                                 " due to rule violations " % src.uuid,
                                 uuid=src.uuid, set_data_errors=e.errors)
    dest.commit()
    return dest


def _copy_data_recursively(src, dest, data, match_uuids):
    dest_children = dest.getChildren()

    if isinstance(dest, RepeatInstanceContainerNode):
        src_rgs = src.get_repeat_groups(dest.getMeta().getName())
        dest_mapping = dict((child.getUUID(), child) for child in dest_children)
        for child_id, src_child in src_rgs.iteritems():
            # FIXME: use uuids here
            dest_child = dest_mapping.pop(src_child.uuid, None)
            if dest_child is None or not match_uuids:
                dest_child = dest.createInstance(src_child.uuid)
            _copy_data_recursively(src_child, dest_child, data,
                                   match_uuids=match_uuids)
        for child_id, dest_child in dest_mapping.iteritems():
            dest.deleteInstance(dest_child)
    else:
        for child in dest_children:
            if child.isLeaf():
                if isinstance(child.getMeta(), (InfoLeaf, SaveLeaf)):
                    continue
                old_value = child.getValue()
                new_value = src.get_leaf(child.getMeta().getName())
                if new_value is not None and new_value != old_value:
                    data[child.getIdentifier()] = new_value
            else:
                _copy_data_recursively(src, child, data,
                                       match_uuids=match_uuids)

