# -*- coding: latin-1 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import os.path
import codecs
import logging

from xml.sax import make_parser
from xml.sax.saxutils import quoteattr, escape

from formed.meta.nodecomponents import RootNode, CaseNode, Leaf, Node
from formed.meta.document import Document
from formed.meta.data import PageNode, LogbookNode, GroupNode, RepeatNode, \
     MatrixNode, RadioNode, BoolLeaf, ChoiceNode, ExternalChoiceListLeaf, \
     IntLeaf, InfoLeaf, TextLeaf, TextAreaLeaf, RuleLeaf, DateLeaf, \
     ConditionalNode, NumericLeaf, ReferenceNode, ExprLeaf, VirtualLeaf, \
     SaveLeaf, FileLeaf, RowNode
from formed.meta.rulegen import RuleGenerator
from formed.meta.io.parser import SAXBuilder
from formed.meta.io.annotations import Annotations
from formed.meta.io.factories import DocumentFactory, SimpleCreator, \
     NodeComponentFactory, NodeFactory, ContentNodeFactory

log = logging.getLogger(__name__)

DOCUMENT_TAG        = u"document"
CASE_TAG            = u"case"
LOGBOOK_TAG         = u"logbook"
PAGE_TAG            = u"page"
GROUP_TAG           = u"group"
ROW_TAG             = u"row"
MATRIX_TAG          = u"radio-matrix"
RADIO_TAG           = u"radio"
BOOL_TAG            = u"bool"
INT_TAG             = u"int"
NUMERIC_TAG         = u"float"
INFO_TAG            = u"info"
SAVE_TAG            = u"save"
TEXT_TAG            = u"text"
CHOICE_TAG          = u"choice"
EXT_CHOICE_LIST_TAG = u"external-choice"
DATE_TAG            = u"date"
TEXT_AREA_TAG       = u"textarea"
RULE_TAG            = u"rule"
CONDITIONAL_TAG     = u"conditional"
REPEAT_TAG          = u"repeat"
REFERENCE_TAG       = u"reference"
EXPR_TAG            = u"expr"
VIRTUAL_TAG         = u"virtual"
FILE_TAG            = u"file"

namesToClasses = {
    DOCUMENT_TAG:        RootNode,
    CASE_TAG:            CaseNode,
    LOGBOOK_TAG:         LogbookNode,
    PAGE_TAG:            PageNode,
    GROUP_TAG:           GroupNode,
    ROW_TAG:             RowNode,
    MATRIX_TAG:          MatrixNode,
    RADIO_TAG:           RadioNode,
    BOOL_TAG:            BoolLeaf,
    CHOICE_TAG:          ChoiceNode,
    EXT_CHOICE_LIST_TAG: ExternalChoiceListLeaf,
    INT_TAG:             IntLeaf,
    NUMERIC_TAG:         NumericLeaf,
    INFO_TAG:            InfoLeaf,
    SAVE_TAG:            SaveLeaf,
    TEXT_TAG:            TextLeaf,
    TEXT_AREA_TAG:       TextAreaLeaf,
    RULE_TAG:            RuleLeaf,
    DATE_TAG:            DateLeaf,
    CONDITIONAL_TAG:     ConditionalNode,
    REPEAT_TAG:          RepeatNode,
    REFERENCE_TAG:       ReferenceNode,
    EXPR_TAG:            ExprLeaf,
    VIRTUAL_TAG:         VirtualLeaf,
    FILE_TAG:            FileLeaf,
    }

factoryCreators = {
    DOCUMENT_TAG:        SimpleCreator(DocumentFactory,      RootNode),
    CASE_TAG:            SimpleCreator(NodeFactory,          CaseNode),
    LOGBOOK_TAG:         SimpleCreator(NodeFactory,          LogbookNode),
    PAGE_TAG:            SimpleCreator(NodeFactory,          PageNode),
    GROUP_TAG:           SimpleCreator(NodeFactory,          GroupNode),
    ROW_TAG:             SimpleCreator(NodeFactory,          RowNode),
    REPEAT_TAG:          SimpleCreator(NodeFactory,          RepeatNode),
    MATRIX_TAG:          SimpleCreator(NodeFactory,          MatrixNode),
    CONDITIONAL_TAG:     SimpleCreator(NodeFactory,          ConditionalNode),
    RADIO_TAG:           SimpleCreator(NodeFactory,          RadioNode),
    CHOICE_TAG:          SimpleCreator(NodeFactory,          ChoiceNode),
    EXT_CHOICE_LIST_TAG: SimpleCreator(NodeComponentFactory,
                                       ExternalChoiceListLeaf),
    BOOL_TAG:            SimpleCreator(NodeComponentFactory, BoolLeaf),
    INT_TAG:             SimpleCreator(NodeComponentFactory, IntLeaf),
    NUMERIC_TAG:         SimpleCreator(NodeComponentFactory, NumericLeaf),
    INFO_TAG:            SimpleCreator(ContentNodeFactory, InfoLeaf),
    SAVE_TAG:            SimpleCreator(NodeComponentFactory, SaveLeaf),
    TEXT_AREA_TAG:       SimpleCreator(NodeComponentFactory, TextAreaLeaf),
    TEXT_TAG:            SimpleCreator(NodeComponentFactory, TextLeaf),
    RULE_TAG:            SimpleCreator(NodeComponentFactory, RuleLeaf),
    REFERENCE_TAG:       SimpleCreator(NodeComponentFactory, ReferenceNode),
    EXPR_TAG:            SimpleCreator(NodeComponentFactory, ExprLeaf),
    VIRTUAL_TAG:         SimpleCreator(NodeComponentFactory, VirtualLeaf),
    DATE_TAG:            SimpleCreator(NodeComponentFactory, DateLeaf),
    FILE_TAG:            SimpleCreator(NodeComponentFactory, FileLeaf),
}

classesToTagNames = dict([(v, k) for k, v in namesToClasses.iteritems()])

ANNOTATIONS_FILE = "annotations.xhtml"

def make_builder(path, documentFactoryCreator):
    builder = SAXBuilder(path)

    if documentFactoryCreator:
        factories = factoryCreators.copy()
        factories[DOCUMENT_TAG] = documentFactoryCreator
    else:
        factories = factoryCreators

    builder.setCreators(factories)

    return builder

def read_document(xml_file, builder, generate_rules, validate_tree,
                  rule_generator=RuleGenerator):
    parser = make_parser()
    parser.setContentHandler(builder)
    parser.parse(xml_file)

    document = builder.getDocument()

    if generate_rules:
        log.debug("generating required rules")
        generator = rule_generator(document)
        generator.generateRequiredRules()
        generator.generateDateSequenceRules()
        document = generator.document

    document.init(validate_tree=validate_tree)

    return document


def openDocument(path, documentFactoryCreator=None, generate_rules=False,
                 validate_tree=True, rule_generator=RuleGenerator):
    """ Load the formed tree from a formed xml specification file.

    path is the full file name of formed xml file.

    documentFactory can be set to pass an own creator object for a formed tree
    document class.

    If required and date rules should be generated automatically on the fly
    generate_rules can be set to True.

    The return value is the formed tree Document.
    """
    # Load annotations first
    dirname = os.path.dirname(path)
    annotations_file = os.path.join(dirname, ANNOTATIONS_FILE)

    annotations = Annotations(annotations_file)

    builder = make_builder(path, documentFactoryCreator)

    f = None
    try:
        f = file(path, "rb")
        document = read_document(f, builder, generate_rules, validate_tree,
                                 rule_generator=rule_generator)
        annotations.decorate(document)
        return document
    finally:
        if f is not None:
            try:
                f.close()
            except:
                pass


def _attributesToXML(nc):
    out = []
    for key, value in sorted(nc.attributes.iteritems()):
        if value:
            out.append(u"%s=%s" % (key, quoteattr(value)))
    x = " ".join(out)
    if x:
        x = " %s" % x
    return x


def _toXML(nc, depth = 0):
    indent = "  " * depth
    if isinstance(nc, Node):
        tagName = classesToTagNames.get(nc.__class__, "node")
        out = [u"%s<%s%s>" % (indent, tagName, _attributesToXML(nc))]
        next_depth = depth + 1
        for child in nc.children:
            out.append(_toXML(child, next_depth))
        out.append(u"%s</%s>" % (indent, tagName))
        return "\n".join(out)
    if isinstance(nc, InfoLeaf):
        tagName = classesToTagNames.get(nc.__class__, "leaf")
        out = [u"%s<%s%s" % (indent, tagName, _attributesToXML(nc))]
        if not nc.children:
            out.append("/>")
        else:
            out.append(">")
            for child in nc.children:
                out.append(_toXML(child, 0))
            out.append(u"</%s>" % (tagName))
        return "".join(out)
    if isinstance(nc, (Leaf, ReferenceNode)):
        tagName = classesToTagNames.get(nc.__class__, "leaf")
        return u"%s<%s%s/>" % (indent, tagName, _attributesToXML(nc))
    if isinstance(nc, Document):
        out = [u'<?xml version="1.0" encoding="UTF-8"?>\n']
        out.append(_toXML(nc.root))
        return "".join(out)
    if isinstance(nc, basestring):
        return escape(nc)
    return ""


def saveDocument(document, path):
    f = None
    try:
        f = codecs.open(path, "wb", "UTF-8")
        print >> f, _toXML(document)
    finally:
        if f is not None:
            try:
                f.close()
            except:
                pass

# vim:set ts=4 sw=4 si et sta sts=4:
