#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2011 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Björn Ricks <bjoern.ricks@intevation.de>
#

import unittest
import datetime

from formed.meta.exprtree import Expr, VariableProvider


class TestRgInstance(object):

    def __init__(self, id, values):
        self.id = id
        self.values = values

    def getIdentifier(self):
        return self.id

    def get_value(self, varname):
        return self.values[varname]


class TestVariableProvider(VariableProvider):

    def __init__(self, vars):
        self.vars = vars
        self.rgs = dict()

    def getValueForVar(self, varname, reference=None):
        if varname in self.vars:
            return (varname, self.vars[varname])
        rg = self.rgs[reference]
        return (varname, rg.get_value(varname))

    def getAllValuesForVar(self, varname):
        return [(varname, value) for value in self.vars[varname]]

    def get_repeat_group_instances(self, name):
        for i, entry in enumerate(self.vars[name]):
            self.rgs[i] = TestRgInstance(i, entry)
        return self.rgs.keys()


class TestExpr(unittest.TestCase):

    def check_expr(self, expr, expected, **variables):
        self.assertEquals(Expr(expr).evaluate(TestVariableProvider(variables)),
                          expected)

    def check_instances(self, expr, expected, **variables):
        self.assertEquals(Expr(expr).evaluate_instances(
                          TestVariableProvider(variables)), expected)

    def test_contains_literals(self):
        self.check_expr("[ 4 8 12 ] 8 contains", True)
        self.check_expr("[ 4 9 16 25 ] 13 contains", False)

    def test_contains_expressions(self):
        self.check_expr("[ 4 2 * 12 ] 8 contains", True)
        self.check_expr("[ 4 6 8 + 15 ] 13 contains", False)

    def test_contains_variables(self):
        self.check_expr("[ 4 $var1 $var2 ] 8 contains", True, var1=5,
                        var2=8)
        self.check_expr("[ 1 2 3 ] $intvar contains", True, intvar=1)
        self.check_expr("[ 0 2 3 ] $intvar contains", False, intvar=1)

    def test_simple_evaluate(self):
        self.check_expr("$intvar 0 ==", False, intvar=1)
        self.check_expr("$intvar 1 ==", True, intvar=1)
        self.check_expr("$intvar 1 +", 2, intvar=1)
        self.check_expr("$intvar 1 -", 0, intvar=1)
        self.check_expr("$intvar 2 *", 4, intvar=2)
        self.check_expr("$intvar 2 /", 1, intvar=2)
        self.check_expr("$intvar 2 %", 1, intvar=3)
        self.check_expr("$intvar 2 %", 0, intvar=2)
        self.check_expr("$intvar 2 **", 4, intvar=2)
        self.check_expr("$intvar 2 **", 4, intvar=2)
        self.check_expr("$intvar abs", 2, intvar=2)
        self.check_expr("$intvar abs", 2, intvar=-2)
        self.check_expr("$intvar $intvar2 min", -2, intvar=2, intvar2=-2)
        self.check_expr("$intvar $intvar2 max", 2, intvar=2, intvar2=-2)
        self.check_expr("$intlist len", 2, intlist=[1, 2])
        self.check_expr("$intlist len 2 ==", True, intlist=[1, 2])
        self.check_expr("$nonevar isset", False, nonevar=None)
        self.check_expr("$intvar isset", True, intvar=1)
        self.check_expr("$intvar $intvar2 >", False, intvar=1, intvar2=2)
        self.check_expr("$intvar2 $intvar >", True, intvar=1, intvar2=2)
        self.check_expr("$intvar2 $intvar <", False, intvar=1, intvar2=2)
        self.check_expr("$intvar $intvar2 <", True, intvar=1, intvar2=2)
        self.check_expr("$intvar str", "1", intvar=1)
        self.check_expr("$intvar str $strvar ==", True, intvar=1, strvar="1")
        self.check_expr("$intvar str $strvar ==", False, intvar=2, strvar="1")
        self.check_expr("$strvar int", 1, strvar="1")
        self.check_expr("$strvar int $intvar ==", True, strvar="1", intvar=1)
        self.check_expr("$strvar int $intvar ==", False, strvar="2", intvar=1)

        self.check_expr("$intvar $intvar2 $falsevar ?", 1, intvar=1, intvar2=2,
                        falsevar=False)
        self.check_expr("$intvar $intvar2 $truevar ?", 2, intvar=1, intvar2=2,
                        truevar=True)

    def test_boolean_evaluate(self):
        self.check_expr("$true $true and", True, true=True)
        self.check_expr("$true isset", True, true=True)
        self.check_expr("$none isset", False, none=None)
        self.check_expr("true", True)
        self.check_expr("true isset", True)
        self.check_expr("$true true ==", True, true=True)
        self.check_expr("$false true ==", False, false=False)
        self.check_expr("$false true !=", True, false=False)
        self.check_expr("false", False)
        self.check_expr("$false false ==", True, false=False)
        self.check_expr("$false true !=", True, false=False)

        try:
            self.check_expr("TRUE", True)
            self.fail("TRUE ist not valid statement for true value")
        except ValueError:
            pass

        try:
            self.check_expr("True", True)
            self.fail("True ist not valid statement for true value")
        except ValueError:
            pass

    def test_list_evaluate(self):
        self.check_expr("%strlist", "world", strlist=["Hello", "world"])
        self.check_expr("%intlist", 2, intlist=[1, 2])
        # stops at first False
        self.check_expr("%boollist", False, boollist=[True, False, True])

    def test_date_evaluate(self):
        date1 = datetime.date(2011, 1, 1)
        date2 = datetime.date(2012, 2, 2)
        date3 = datetime.date(2012, 2, 3)
        date4 = datetime.date(2011, 12, 31)
        self.check_expr("$date1 $date2 years", 1, date1=date1, date2=date2)
        self.check_expr("$date2 $date1 years", 1, date1=date1, date2=date2)
        self.check_expr("$date1 $date2 months", 13, date1=date1, date2=date2)
        self.check_expr("$date2 $date1 months", 13, date1=date1, date2=date2)
        self.check_expr("$date1 $date2 months", 1, date1=date3, date2=date4)
        self.check_expr("1 1 2003 date 31 1 2002 date months", 11)
        self.check_expr("31 1 2002 date 1 2 2002 date months", 0)
        self.check_expr("$date1 $date2 days", 397, date1=date1, date2=date2)
        self.check_expr("$date2 $date3 days", 1, date2=date2, date3=date3)
        self.check_expr("$date3 $date2 days", 1, date2=date2, date3=date3)
        self.check_expr("24 01 2011 date $date1 days", 23, date1=date1)

    def test_types(self):
        self.check_expr("'abc'", "abc")
        self.check_expr('"abc"', "abc")
        self.check_expr('"hello world"', "hello world")
        self.check_expr("'hello world'", "hello world")
        try:
            self.check_expr("abc", "abc")
            self.fail("abc is not a valid expression")
        except ValueError:
            pass
        self.check_expr("''", "")
        self.check_expr("\"\"", "")

        self.check_expr("1", 1)
        self.check_expr("01", 1)

        self.check_expr("1.0", 1.0)
        self.check_expr("01.0", 1.0)
        self.check_expr("1.0000", 1.0)

        self.check_expr("01 01 2011 date", datetime.date(2011, 1, 1))
        self.check_expr("1 1 2011 date", datetime.date(2011, 1, 1))
        self.check_expr("01 1 2011 date", datetime.date(2011, 1, 1))

        # check list inclusive list order
        self.check_expr("[ 1 2 ]", [1, 2])
        try:
            self.check_expr("[ a b ]", ["a", "b"])
            self.fail("[ a b ] is not a valid expression")
        except ValueError:
            pass
        self.check_expr("[ 'a' 'b' ]", ["a", "b"])
        self.check_expr("[ \"a\" \"b\" ]", ["a", "b"])

    def test_numeric_operator(self):
        self.check_expr("1 1 +", 2)
        self.check_expr("1 1 *", 1)

        self.check_expr("1.0 1.0 +", 2.0)
        self.check_expr("1.0 1.0 *", 1.0)

        try:
            self.check_expr("[ 1 2 ] [ 3 4 ] +", False)
            self.fail("+ operator is not valid for lists")
        except ValueError:
            pass

        try:
            self.check_expr("'a' 'b' +", False)
            self.fail("+ operator is not valid for strings")
        except ValueError:
            pass

        try:
            self.check_expr("[ 1 2 ] [ 3 4 ] *", False)
            self.fail("* operator is not valid for lists")
        except ValueError:
            pass

        try:
            self.check_expr("'a' 'b' *", False)
            self.fail("* operator is not valid for strings")
        except ValueError:
            pass

    def test_sum(self):
        self.check_expr("[ 1 1 1 ] sum", 3)
        self.check_expr("$list sum", 3, list=[1, 1, 1])
        try:
            self.check_expr("1 sum", False)
            self.fail("sum operator should only be usable with lists")
        except ValueError:
            pass

    def test_evaluate_instances(self):
        self.check_instances("$intvar", [(1, {u'intvar': 1})], intvar=1)
        self.check_instances("%strlist", [('a', {u'strlist': 'a'}),
                                          ('b', {u'strlist': 'b'})],
                             strlist=["a", "b"])
        self.check_instances("%boollist", [
                             (True, {u'boollist': True}),
                             (False, {u'boollist': False}),
                             (True, {u'boollist': True})],
                             boollist=[True, False, True])

    def test_rgmap(self):
        self.check_expr("$rgvar \"rginst\" rgmap", [1, 2],
                        rginst=[{"rgvar": 1}, {"rgvar": 2}])
        self.check_expr("$rgvar isset \"rginst\" rgmap", [False, True],
                        rginst=[{"rgvar": None}, {"rgvar": 2}])
        self.check_expr("$rgvar1 $rgvar2 + \"rginst\" rgmap", [3, 3],
                        rginst=[{"rgvar1": 1, "rgvar2": 2},
                                {"rgvar1": 2, "rgvar2": 1}])
        self.check_expr("$rgvar1 $rgvar2 + \"rginst\" rgmap sum", 6,
                        rginst=[{"rgvar1": 1, "rgvar2": 2},
                                {"rgvar1": 2, "rgvar2": 1}])
        self.check_expr("$rgvar1 $rgvar2 $var + + \"rginst\" rgmap sum", 8,
                        rginst=[{"rgvar1": 1, "rgvar2": 2},
                                {"rgvar1": 2, "rgvar2": 1}],
                        var=1)

    def test_coalesce(self):
        self.check_expr("$var1 $var2 coalesce", 1, var1=1, var2=2)
        self.check_expr("$var1 $var2 coalesce", 2, var1=None, var2=2)

    def test_least_nonnull(self):
        self.check_expr("$start $end least", 10, start=10, end=20)

    def test_least_firstnull(self):
        self.check_expr("$start $end least", 20, start=None, end=20)

    def test_least_secondnull(self):
        self.check_expr("$start $end least", 10, start=10, end=None)

    def test_least_bothnull(self):
        self.check_expr("$start $end least", None, start=None, end=None)

    def test_greatest_nonnull(self):
        self.check_expr("$start $end greatest", 20, start=10, end=20)

    def test_greatest_firstnull(self):
        self.check_expr("$start $end greatest", 20, start=None, end=20)

    def test_greatest_secondnull(self):
        self.check_expr("$start $end greatest", 10, start=10, end=None)

    def test_greatest_bothnull(self):
        self.check_expr("$start $end greatest", None, start=None, end=None)


if __name__ == "__main__":
    unittest.main()
