# -*- coding: utf-8 -*-
#
# (c) 2008 by Intevation GmbH
# This is Free software under the GPLv3. 
# See LICENSE comming with the source of 'mpuls'
# for details.
#
# author: Torsten Irländer <torsten.irlaender@intevation.de>


import sys, logging, string
import datetime 
import decimal
import traceback
from libmpuls import LIBMPULS_VERSION
from config import TMP_TABLE, EVAL_TBL, TMP_SET, TMP_EVAL, PREP_STMT, \
SET_FILTER_STMNT, EVAL_FILTER_STMNT, SET_SUB_STMNT, TMP_SUB

class NullHandler(logging.Handler):
    def emit(self, record):
        pass

h = NullHandler()
logging.basicConfig()
log = logging.getLogger(__name__)
log.setLevel(logging.ERROR)

def get_prepare_query(baseq, table, where, xtable, xfield, xjoin, ytable, yfield, yjoin, function=None):
    '''
    Returns the SQL-String to prepare the data for further evaluation.
    The preparations is written in a temporary table which will be further used.

    @param id: Id of the evaluation 
    @type id: Integer
    @return: preparation sql for the evaluation
    @rtype: String
    '''

    # Build basequery
    bq = baseq % {'xfield': xfield,
                  'yfield': yfield,
                  'from': table,
                  'where': where}

    # Build prepare query 
    pq = PREP_STMT % {'tmp_tbl': TMP_TABLE, 
                      'basequery': bq,
                      'xtable': xtable,
                      'ytable': ytable, 
                      'xjoinon': xjoin,
                      'yjoinon': yjoin,
                      'function': function or 'count(m.id)'}
    return pq

def get_minphasestmnt(phases):
    '''Returns the phase-part of the minimum required phase depending of
    notation. If only one phase is given, then return old 'phase >=' statement.
    Else return 'phase in (x,y,z)'. Parameter phases is a comma seperated
    list'''
    out = []
    tok = phases.split(',')
    if len(tok) == 1:
        out.append('phase >= %s' % tok[0])
    else:
        out.append('phase in ( %s )' % ",".join(tok))
    return " ".join(out)

class EvaluationSet:

    '''
    The evaluationset is a bundle of one or more single evaluations, which can
    be performed at once. The set has some free parameters which can be
    configured in a configuration file. These params are:

     - Id
     - Name
     - Description

     - Start and Enddate
     - Precodition or Requirement

    Precondition and Requirents
    ===========================
    You can define some precoditions or requirements which need to be met by the
    cases before they are considered to be part of the evaluation set. As
    result all evalutions within the set are performed on a subset of cases
    which met the defined precoditions or requirements.

    Requirements
    ------------
    Requirments can be defined based the phase or validiy of each case.

     - Phase: The phase is an indicator for the fillstate of a case. So if some
       evaluations needs some data to be present, the phase indicates if there is
       a chance that the user already has provided the information. Only cases
       which have the minimum phase are considered for evaluation.
     - Validity: Some fields a crucial for the evaluation. The validty guarantees
       that the cases will met all minimum requirments for the evalution. The
       cases can have different categorys of validity. Only cases which have all
       categories of validity will be taken for evaluation.

    Preconditions
    -------------
    Precoditions are userdefined conditions which needs to be met be the case to
    be part of the evaluation.

     - Dates: Only cases which start and enddate interfers with the start and
     enddate of the evaluation start and enddate will be taken for evaluation.
    '''

    def __init__(self, config, verbose=False):
        self.config     = config # EvaluationConfig-object 
        self.conn       = self.config.conn # DB-Connection
        self.id         = self.config.get_setid()
        self.mode       = self.config.get_setmode()
        self.name       = self.config.get_setname()
        self.desc       = self.config.get_setdesc()
        self.agency     = self.config.get_agency()
        self.enabled    = self.config.elist 
        self.evals      = []     # List of evaluations

        if verbose:
            log.setLevel(logging.DEBUG)

        # Requirements
        self.rphase         = self.config.get_required_setphase()
        self.rvalidity      = self.config.get_required_setvalidity()
        self.num_total      = 0     # Number of total cases in the database
        self.num_valid      = 0     # Number of cases which met the basic requirements 
        self.num_considered = 0 # Number of cases which are the base for
                                # evaluation (met also user defined filters
                                # like dates etc.))

        # Precodnditons
        self.sdate      = self.config.sdate 
        self.edate      = self.config.edate 
        self.sdatefield = self.config.sdatefield
        self.edatefield = self.config.edatefield
        self.pphase     = self.config.pphase
        self.sqlwhere   = self.config.sqlwhere
        
        # Human readable filters
        self.filters    = self.config.evaluation_filter

        # Version info. These variables may contain information on which
        # versions of the different components where used during the
        # evaluation.
        try:
            cur = self.conn.cursor()
            cur.execute("SELECT value FROM dbinfo WHERE key = 'version'")
            self.v_database = cur.fetchone()[0]
            cur.execute("SELECT value FROM dbinfo WHERE key = 'datadump'")
            self.v_dataset = cur.fetchone()[0]
        except StandardError, err:
            log.error("reading dbinfo: %s" % err)
            self.v_dataset    = None
            self.v_database   = None

        self.v_prequery   = self.config.sql
        self.v_libmpuls   = LIBMPULS_VERSION 
        self.v_evalconfig = self.config.get_version()

        for id in self.config.get_evaluations():
            name   = self.config.get_name(id)
            desc   = self.config.get_desc(id)
            self.evals.append(Evaluation(id, config, name, desc))

    def _filter(self, conn):
        '''Will create a tmp table for evaluation (set wide). This table only
        contains cases which match the set requiremends like 1. phases, 2.
        validity, 3. startdate and enddate, 4. userdefined subselect
        (optional)'''
        tags, phase, dates, pphase, sqlwhere = ('TRUE', 'TRUE', 'TRUE', 'TRUE', 'TRUE')

        # Requirements (hardcoded in evalconfig)
        if self.rvalidity:
            tags = """EXISTS (SELECT t.case_id from nm_tags_cases_view t WHERE t.case_id = m.id AND t.tag_id in (%s))""" % ",".join(self.rvalidity)
        if self.rphase:
            phase = get_minphasestmnt(self.rphase)

        req_where = " AND ".join([phase]) 


        # Preconditions (userdefined)
        # TODO: I think pphase can be removed as it is not usefull anyway.
        # minimum phases are set in each evaluation and i can not think of a
        # reason why to set an additional minimum phase
        if self.pphase:
            pphase = get_minphasestmnt(self.pphase)
        if self.sqlwhere:
            sqlwhere = "( %s )" % self.sqlwhere
        if None not in [self.sdate, self.sdatefield, self.edate, self.edatefield] and \
           'None' not in [self.sdate, self.sdatefield, self.edate, self.edatefield] and \
           self.config.adele:
               
            dates = "'%s'::date <= coalesce(%s, now()) AND '%s'::date >= %s" % (
                    self.sdate,
                    self.edatefield,
                    self.edate,
                    self.sdatefield)
        else:
            dates = "TRUE"

        log.debug("my dates: %s" % dates)

        where = " AND ".join([phase, dates, pphase, sqlwhere]) 
        log.debug("my where: %s" % where)

        try:
            try:
                # 0. Build subquery. The subquery can be defined by the user
                # and is a complete sql select statement. 
                # e.g SELECT # %(fields)s from master_tbl
                # Further store subquery in e temporary subquery table
                cur = conn.cursor()

                if self.config.sql: subquery = self.config.sql % {'fields': 'm.*'}
                else: subquery = "SELECT * from %s" % EVAL_TBL
                query = SET_SUB_STMNT % {'tmp_tbl': TMP_SUB, 'subquery': subquery}
                if self.config.sql and self.config.sqlparams:
                    cur.execute(query, self.config.sqlparams)
                else:
                    cur.execute(query)
                log.debug("Save subquery in temp table: %s" % cur.query)

                # 1. Count how many total cases are in the db
                cur.execute('SELECT count(m.id) FROM %s m' % TMP_SUB)
                self.num_total = int(cur.fetchone()[0])
                log.debug("Count total cases: %s" % cur.query)

                # 2. Filter valid cases from subquery Valid means havin
                # required validity and phases
                d = {'tmp_tbl':TMP_SET, 'tags': tags, 'where': where, 'subquery': TMP_SUB}
                cur.execute(SET_FILTER_STMNT % d)
                log.debug("Set set: %s" % cur.query)

                # 3. Count how many valid cases are in the db. 
                cur.execute('SELECT count(m.id) FROM %s m' % TMP_SET)
                log.debug("Count valid cases: %s" % cur.query)
                self.num_valid = int(cur.fetchone()[0])
                self.num_considered = self.num_valid
            except StandardError, err:
                log.error("filter set: %s %s "  % (err, cur.query))
        finally: cur = None

    def evaluate(self):
        conn, cur = None, None
        try:
            try:
                conn = self.conn
                log.debug("-- START FILTER SET")
                self._filter(self.conn)
                log.debug("-- END FILTER SET")
                for eval in self.evals:
                    eval.perform(self.conn)
            except Exception, err:
                log.error('Error on evaluation set: %s' % err)
        finally:
            pass

    def export(self, exporter):
        return exporter.export(self)

class Evaluation:

    '''
    A single evaluation. The evaluation consists of one or many Evaluationsquerys. The result is a 2 dimensional crosstabluar.
    '''

    def __init__(self, id, config=None, name=None, desc=None):
        self.id        = id
        self.config    = config
        self.name      = name
        self.desc      = desc
        self.short_name = self.config.get_short_name(id)
        self.bq        = self.config.get_basequery(id)
        self.table     = self.config.get_basetable(id)
        self.where     = self.config.get_basewhere(id)
        self.aggregate = self.config.get_baseaggregate(id)
        self.aggregate_total = self.config.get_baseaggregatetotal(id)


        if not self.aggregate_total:
            # Hackish. Find better way to determine how to sum up the results.
            if self.aggregate.find('sum') > 0:
                self.aggregate_total = "avg"
            else:
                self.aggregate_total = "sum"

        self.rphase    = self.config.get_required_evalphase(id)
        self.rvalidity = self.config.get_required_evalvalidity(id)
        self.num_total      = 0
        self.num_valid      = 0
        self.num_considered = 0

        # Depending on how many dimensions are defined the evaluation holds
        # several singe Querys in a multidimensional array. Each query will
        # create a single resultset. At the end those resultsets need to be
        # combined.
        #         xdim 
        #    +------+------+
        # y  |  Q1  |  Q2  | 
        # d  +------+------+
        # i  |  Q3  |  Q4  | 
        # m  +------+------+

        self.queries = []
        for ydim in self.config.get_ydims(id):
            ytable = self.config.get_table(ydim)
            ylabel = self.config.get_label(ydim)
            yfield = self.config.get_field(ydim)
            yjoin  = self.config.get_join(ydim)
            ywhere = self.config.get_where(ydim)
            for xdim in self.config.get_xdims(id):
                xtable = self.config.get_table(xdim)
                xlabel = self.config.get_label(xdim)
                xfield = self.config.get_field(xdim)
                xjoin  = self.config.get_join(xdim)
                mode = self.config.get_setmode()
                if not mode:
                    mode = self.config.get_evaluation_mode(id)
                self.queries.append(EvaluationQuery(
                    get_prepare_query(self.bq, 
                        self.table, 
                        " AND ".join([self.where, ywhere]), 
                        xtable, 
                        xfield, 
                        xjoin, 
                        ytable, 
                        yfield, 
                        yjoin,
                        self.aggregate), 
                        xlabel, 
                        ylabel,
                        self.config.sdate,
                        self.config.edate,
                        self.aggregate_total, 
                        mode))

        log.info('Initialised %s evaluation' % len(self.queries))

    def __str__(self):
        return self.export(EvaluationExportConsole())

    def _filter(self, conn):
        tags, phase = ('TRUE', 'TRUE')

        # Requirements
        if self.rvalidity:
            tags = """EXISTS (SELECT t.case_id from nm_tags_cases_view t WHERE t.case_id = m.id AND t.tag_id in (%s))""" % ",".join(self.rvalidity)
        if self.rphase:
            phase = get_minphasestmnt(self.rphase)

        where = phase
        try:
            try:
                # 1. Store the valid cases in a temp evaluation table,
                # but also add date restrictions or user defined sql 
                cur = conn.cursor()
                cur.execute(EVAL_FILTER_STMNT % {'tmp_tbl':TMP_EVAL, 'tags': tags, 'where': where})
                log.debug(cur.query)
                cur = conn.cursor()
                
                # 2. Count how man cases where in the evaluationset table
                # (maximum amount of cases for the evaluations)) 
                cur.execute('SELECT count(id) FROM %s' % TMP_SET)
                self.num_total = int(cur.fetchone()[0])

                # 3. Count how many cases are considred for this evaluation
                cur.execute('SELECT count(id) FROM %s' % TMP_EVAL)
                self.num_considered = int(cur.fetchone()[0])
                self.num_valid = self.num_considered
            except StandardError, err:
                log.error("filter eval: %s %s"  % (err, cur.query))
        finally: 
            cur = None

    def perform(self, conn):
        log.debug("-- START EVAL ID: %s" % self.id)
        self._filter(conn)
        cur = conn.cursor()
        log.debug('Creating index on tmp_eval on ID')
        cur.execute('CREATE INDEX tmp_id_index ON tmp_eval (id);')
        log.debug("-- END EVAL ID: %s" % self.id)
        for num, q in enumerate(self.queries):
            log.debug("-- START QUERY %s" % num)
            q.perform(conn)
            log.debug("-- END QUERY %s" % num)

        # Drop tmp table as it needs to be new for each evaluation
        try:
            try:
                cur = conn.cursor()
                cur.execute('DROP TABLE %s' % TMP_EVAL)
            except:
                log.error('Error while dropping table: %s' % TMP_EVAL)
        finally:
            cur = None

    def get_result(self):
        return self.queries

class EvaluationQuery:

    def __init__(self, pq=None, xlabel='X-Label', ylabel='Y-Label', sdate=None, edate=None, total="sum", mode=None):
        self.pq     = pq
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.result = {}
        self.sdate  = sdate
        self.edate  = edate
        self.total  = total 
        self.mode   = mode # Display may vary depending on mode. e.g trendanalyse 
        log.info('Mode: %s' % self.mode)

    def perform(self, conn):
        cur = conn.cursor()
        self.result = []
        for sdate, edate, name in self._get_startenddates(
                self.sdate, self.edate, self.mode == 'trend'):
            try:
                try:
                    log.debug("-- PREPARE")
                    yfields = self._prepare(conn, sdate, edate)
                    log.debug("-- FINAL")
                    test = self._final(conn, yfields, sdate, edate, name)
                    self.result.extend(test)
                except StandardError, err:
                    log.error("perform: %s "  % err)
            finally:
                cur.execute('DROP TABLE %s' % TMP_TABLE)
                log.debug(cur.query)
        return self.result

    def _prepare(self, conn, sdate, edate):
        yfields = []
        try:
            try:

                # Evalutions are done internal on subsets (temporary table) of
                # the
                # origin data. So replace the orign table name with the name
                # of the name of the temporary table.
                pq  = string.replace(self.pq, EVAL_TBL, TMP_EVAL)

                cur = conn.cursor()
                cur.execute(pq % {'sdate': sdate, 'edate': edate})
                log.debug(cur.query)
                cur.execute('SELECT DISTINCT xid, x FROM %s ORDER BY xid' % TMP_TABLE)
                log.debug(cur.query)
                for row in cur.fetchall():
                    yfields.append((row[0], row[1]))
                return yfields
            except StandardError, err:
                log.error("prepare: %s %s "  % (err, cur.query))
                raise
        finally: cur = None

    def _sum(self, result):
        '''Sums up all rows to a total row and add it at the end of the result
        if the input result has more than on row'''

        total = []
        for row in result:
            for cols in row.itervalues():
                for key, value in cols:
                    total.append((key, 0))
            break;

        for row in result:
            for cols in row.itervalues():
                for num, col in enumerate(cols):
                    key, value = total[num]
                    if col[1] != None:
                        value += decimal.Decimal(col[1])
                    total[num] = (key, value)

        if len(result) > 1: 
            result.append({'total':total})
        return result


    def _get_startenddates(self, sdate, edate, trend):
        def _get_next_month(y, m ,d=1):
            if m < 12:
                return y,m+1,d
            else:
                return y+1,1,d

        def _get_day_before(y, m, d):
            d = datetime.date(y,m,d)
            d = d - datetime.timedelta(days=1)
            return d.year, d.month, d.day

        dates = []
        if trend:
            y,m,d = [int(x) for x in sdate.split('-')]
            sy,sm,sd = [int(x) for x in sdate.split('-')]
            ey,em,ed = [int(x) for x in edate.split('-')]
            test = True
            while test:
                # Build dates
                msdate = "%s-%s-%s" % (sy, sm, sd)
                sd = 1 # Next month will begin on first
                if sy == ey and sm == em:
                    medate = "%s-%s-%s" % (ey, em, ed)
                    test = False
                else:
                    dy, dm, dd = _get_next_month(sy, sm)
                    medate = "%s-%s-%s" % (_get_day_before(dy, dm, dd))
                name = "%s.%s" % (sm,sy)
                dates.append((msdate, medate, name))
                sm+=1
                if sm > 12:
                    sm = 1
                    sy+=1
        else:
            name = None
            dates.append((sdate, edate, name))
        return dates

    def _final(self, conn, fields, sdate=None, edate=None, name=None):
        result = []
        try:
            try:
                sql = self._build(conn, fields)
                cur = conn.cursor()
                cur.execute(sql % {'sdate': sdate, 'edate': edate})
                fieldnames = cur.description
                for row in cur.fetchall():
                    rrow = {}
                    if name:
                        label = "%s %s" % (row[0], name)
                    else:
                        label = row[0]
                    rrow[label] = []
                    for id in range(1, len(row)):
                        value = row[id]
                        key = fieldnames[id][0]
                        rrow[label].append((key, value))
                    result.append(rrow)
                return self._sum(result)
            except StandardError, err:
                log.error("final: %s %s"  % (err, cur.query))
        finally:
            cur = None

    def _build(self, conn, fields):
        sql = ['SELECT']
        fields_buf = ['y']
        num_total = 0
        # TODO: If there no cases in tmp table because of filtering invalid cases. The result table is broken. 
        cur = conn.cursor()
        for field in fields:
            cur.execute('select sum(count) from %s where xid = %s' % (TMP_TABLE,
                        field[0]))
            num_cases = cur.fetchone()[0]
            num_total += num_cases

            if self.total == 'avg' and num_cases > 0:
                fields_buf.append(
                    'SUM(CASE WHEN xid=%s THEN result ELSE 0 END)/%s as "%s"'
                    % (field[0], num_cases, field[1]))
            else:
                fields_buf.append(
                    'SUM(CASE WHEN xid=%s THEN result ELSE 0 END) as "%s"' 
                    % (field[0], field[1]))

        if self.total == 'sum':
            fields_buf.append('SUM(result) as gesamt')

        elif self.total == 'avg':
            if num_total > 0:
                fields_buf.append('SUM(result)/%s as gesamt' % num_total)
            else:
                fields_buf.append('0 as gesamt')

        sql.append(",".join(fields_buf))
        sql.append('FROM %s GROUP BY y, yid ORDER BY yid' % TMP_TABLE)
        return "\n".join(sql)

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8 :
